import os
from pathlib import Path
from warnings import warn

import onetick.py as otp


def __create_bat_script_to_run_config(dir_path, config_path):
    with open(os.path.join(dir_path, "run.bat"), "w") as fout:
        fout.write(f"set ONE_TICK_CONFIG={config_path}\n")

        path = os.path.join(omd_dist_path(), "one_tick", "bin", "OneTickDisplay.exe")

        fout.write(f'start "OneTick GUI" {path} -context DEFAULT\n')


def __create_bat_script_to_run_query_designer(dir_path, config_path):
    with open(os.path.join(dir_path, "run_query_designer.bat"), "w") as fout:
        fout.write(f"set ONE_TICK_CONFIG={config_path}\n")

        path = os.path.join(omd_dist_path(), "one_tick", "bin", "QueryDesigner.exe")

        fout.write(f'start "Query Designer" {path} -context DEFAULT\n')


def write_config_for_tick_server_run(fout, config_path, path):
    fout.write('if [ "$1" == "-" ]; then\n')
    fout.write(
        " " * 4
        + "onetick server --cfg "
        + config_path
        + " -n "
        + os.path.basename(config_path).split(".")[0]
        + " ${@:2}\n"  # this option allows to propagate other custom parameters into the script externally
    )
    fout.write("else\n")
    fout.write(" " * 4 + path + " -context DEFAULT -port $MAIN_TS_PORT > /dev/null 2>&1\n")
    fout.write("fi\n")


def __create_shell_script_to_run_config(dir_path, config_path):
    file_name = os.path.join(dir_path, "run.sh")

    path = os.path.join(omd_dist_path(), "one_tick", "bin", "tick_server.exe")

    with open(file_name, "w") as fout:
        fout.write("#!/bin/bash\n")
        fout.write('if [[ -z "$1" && -z "$MAIN_TS_PORT" ]]; then\n')
        fout.write(
            " " * 4
            + 'echo "Please, define port either using MAIN_TS_PORT environment'
              ' variable or as a first parameter to the ./run.sh script"\n'
        )
        fout.write(" " * 4 + "exit 1\n")
        fout.write("\n")
        fout.write("fi\n")
        fout.write('if [[ -n "$1" ]]; then\n')
        fout.write(" " * 4 + "MAIN_TS_PORT=$1\n")
        fout.write("fi\n")
        fout.write("\n")
        fout.write(f"export ONE_TICK_CONFIG={config_path}\n")
        write_config_for_tick_server_run(fout, config_path, path)

    os.chmod(file_name, 0o750)


def __create_shell_script_to_run_dashboard(dir_path, config_path, main_install_dir):
    if main_install_dir is None:
        return
    file_name = os.path.join(dir_path, "run_dashboard.sh")
    path_ts = os.path.join(omd_dist_path(), "one_tick", "bin", "tick_server.exe")
    client_data_path = os.path.join(main_install_dir, "client_data")
    common_path = os.path.join(main_install_dir, "sol", "common")
    tomcat_cfg = os.path.join(dir_path, 'tomcat.txt')
    db_locator_remote = os.path.join(common_path, "config", "locator.remote")

    Path(tomcat_cfg).write_text('')

    with open(config_path, 'a') as f:
        f.write(f'\nDB_LOCATOR.REMOTE = "{db_locator_remote}"')

    with open(file_name, "w") as fout:
        fout.write("#!/bin/bash\n")
        fout.write(
            'if [[ -z "$1" &&'
            ' -z "$MAIN_TS_PORT" &&'
            ' -z "$CATALINA_HTTP_PORT" &&'
            ' -z "$CATALINA_SERVER_PORT" ]]; then\n')
        fout.write(
            " " * 4
            + 'echo "Please, define ts port to use sequentially ports for catalina '
              'or using \\$MAIN_TS_PORT, \\$CATALINA_HTTP_PORT, '
              '\\$CATALINA_SERVER_PORT environment variable or '
              'as a first parameter to the script"\n'
        )
        fout.write(" " * 4 + "exit 1\n")
        fout.write("\n")
        fout.write("fi\n")
        fout.write('if [[ -n "$1" ]]; then\n')
        fout.write(" " * 4 + "export CATALINA_HTTP_PORT=$1\n")
        fout.write(" " * 4 + "export MAIN_TS_PORT=$(($1 + 1))\n")
        fout.write(" " * 4 + "export CATALINA_SERVER_PORT=$(($1 + 2))\n")
        fout.write("fi\n")
        fout.write("\n")
        fout.write('echo "Running dashboard with CATALINA_HTTP_PORT=$CATALINA_HTTP_PORT, '
                   'MAIN_TS_PORT=$MAIN_TS_PORT, '
                   'CATALINA_SERVER_PORT=${CATALINA_SERVER_PORT}"\n')
        fout.write("export ONE_TICK_CONFIG=" + config_path + "\n")
        fout.write("export MAIN_CLIENT_DIR=" + client_data_path + "\n")
        fout.write("export MAIN_INSTALL_DIR=" + main_install_dir + "\n")
        fout.write("export OMD_WEB_DASHBOARD_CONFIG_FILE="
                   + os.path.join(client_data_path, "config", "web_dashboard.cfg") + "\n")
        fout.write("export COMMON_PACK=" + common_path + "\n")
        fout.write("export SURV_PACK=" + os.path.join(main_install_dir,
                                                      "sol", "surveillance") + "\n")
        fout.write("export BESTEX_PACK=" + os.path.join(main_install_dir,
                                                        "sol", "bestex") + "\n")
        fout.write("export WORKFLOW_PACK=" + os.path.join(main_install_dir,
                                                          "sol", "workflow") + "\n")
        fout.write("source " + os.path.join(common_path, "config", "src",
                                            "env_common_functions.sh") + "\n")

        fout.write('if isAWSHosted\n')
        fout.write('then\n')
        fout.write('  export_if_empty MAIN_ONE_TICK_DIR="/opt/one_market_data/one_tick"\n')
        fout.write('  export_if_empty LICENSE_DIR="/license"\n')
        fout.write('  export_if_empty VENDOR_DATA_DIR="${LOCATOR_LOCAL_DATA_DIR}/vendor_data/"\n')
        fout.write('  export_if_empty JAVA_HOME="/usr/bin/java"\n')
        fout.write('\n')
        fout.write('  if [[ -d /usr/share/tomcat8 ]]\n')
        fout.write('  then\n')
        fout.write('    export_if_empty CATALINA_HOME="/usr/share/tomcat8"\n')
        fout.write('  else\n')
        fout.write('    export_if_empty CATALINA_HOME="/usr/share/tomcat"\n')
        fout.write('  fi\n')
        fout.write('\n')
        fout.write('  # trying system tomcat, then client tomcat, then sol tomcat\n')
        fout.write('  if [[ -d "/etc/omd/tomcat" ]];  then\n')
        fout.write('    export_if_empty CATALINA_BASE="/etc/omd/tomcat"\n')
        fout.write('  elif [[ -d "${MAIN_CLIENT_DIR}/config/tomcat" ]]; then\n')
        fout.write('    export_if_empty CATALINA_BASE="${MAIN_CLIENT_DIR}/config/tomcat"\n')
        fout.write('  else\n')
        fout.write('    export_if_empty CATALINA_BASE="${COMMON_PACK}/tomcat"\n')
        fout.write('  fi\n')
        fout.write('fi\n')
        fout.write('\n')
        fout.write('if isDeployed\n')
        fout.write('then\n')
        fout.write('  export_if_empty MAIN_TS_PORT=12345\n')
        fout.write('  export_if_empty CATALINA_SERVER_PORT=12346\n')
        fout.write('  export_if_empty CATALINA_HTTP_PORT=8080\n')
        fout.write('  export_if_empty OMD_WEB_DASHBOARD_CONFIG_FILE=${MAIN_CLIENT_DIR}/config/web_dashboard.cfg\n')
        fout.write('  export_if_empty MAIN_ONE_TICK_DIR="$MAIN_INSTALL_DIR/../onetick/one_market_data/one_tick"\n')
        fout.write('  export_if_empty LICENSE_DIR="$MAIN_INSTALL_DIR/../license"\n')
        fout.write('  export_if_empty JAVA_HOME="$MAIN_INSTALL_DIR/../java/jdk1.8.0_202"\n')
        fout.write('  export_if_empty CATALINA_HOME="$MAIN_INSTALL_DIR/../tomcat"\n')
        fout.write('  export_if_empty CATALINA_BASE="$MAIN_CLIENT_DIR/config/tomcat"\n')
        fout.write('fi\n')
        fout.write('export_if_empty CATALINA_PID="$CATALINA_BASE/bin/pid"\n')
        fout.write('export_if_empty CATALINA_TMPDIR="$CATALINA_BASE/temp"\n')
        fout.write(f'export TOMCAT_CFG="{tomcat_cfg}"\n')
        fout.write(
            'export JAVA_OPTS="-Djava.awt.headless=true -Dfile.encoding=UTF-8 '
            '-server -Xms512m -Xmx512m -XX:+UseParNewGC -XX:+UseConcMarkSweepGC '
            '-XX:+AggressiveOpts -DServerPort=$CATALINA_SERVER_PORT $JAVA_OPTS"\n')
        fout.write(
            'export CATALINA_OPTS="-Djna.tmpdir=$CATALINA_TMPDIR '
            '-Dcatalina.log.path=$MAIN_LOG_DIR -DHttpPort=$CATALINA_HTTP_PORT '
            '-DHttpsPort=$CATALINA_HTTPS_PORT -DUser=$USER $CATALINA_OPTS"\n')
        fout.write(
            'export PATH="$MAIN_ONE_TICK_DIR/bin:$JAVA_HOME/bin:$CATALINA_HOME/bin:$COMMON_PACK/scripts:$PATH"\n')
        fout.write(
            'export LD_LIBRARY_PATH="$MAIN_ONE_TICK_DIR/bin:'
            '$MAIN_ONE_TICK_DIR/lib:$MAIN_ONE_TICK_DIR/bin/wombat:'
            '$MAIN_ONE_TICK_DIR/bin/activ:$MAIN_ONE_TICK_DIR/bin/rfa:'
            '$MAIN_ONE_TICK_DIR/bin/spryware:$LD_LIBRARY_PATH"\n')
        fout.write('export_if_empty MAIN_TS_IP="localhost"\n')
        fout.write('if isAWSHosted\n')
        fout.write('then\n')
        fout.write('  ' + os.path.join(common_path, "scripts",
                                       "impl", "start_apache_aws.sh") + "\n")
        fout.write('fi\n')
        fout.write('if isDeployed\n')
        fout.write('then\n')
        fout.write('  ' + os.path.join(common_path, "scripts",
                                       "impl", "start_apache.sh") + "\n")
        fout.write('fi\n')
        write_config_for_tick_server_run(fout, config_path, path_ts)
        fout.write('if isAWSHosted\n')
        fout.write('then\n')
        fout.write('  ' + os.path.join(common_path, "scripts",
                                       "impl", "stop_apache_aws.sh") + "\n")
        fout.write('fi\n')
        fout.write('if isDeployed\n')
        fout.write('then\n')
        fout.write('  ' + os.path.join(common_path, "scripts",
                                       "impl", "stop_apache.sh") + "\n")
        fout.write('fi\n')

    os.chmod(file_name, 0o750)


def create_file_to_run_config(dir_path, config_path, main_install_dir=None):
    if otp.__webapi__:
        warn("OTP_WEBAPI is set, means local configs will not work with remote OneTick server")
    if os.name == "nt":
        __create_bat_script_to_run_config(dir_path, config_path)
        __create_bat_script_to_run_query_designer(dir_path, config_path)
    else:
        __create_shell_script_to_run_config(dir_path, config_path)
        __create_shell_script_to_run_dashboard(dir_path, config_path,
                                               main_install_dir)


def omd_dist_path():
    """
    Returns path to the OneTick distribution calculated from the
    PYTHONPATH or MAIN_ONE_TICK_DIR. If path is not found, then FileNotFoundError is risen.
    Example for a linux system: /opt/one_market_data
    """
    omd_prefix = Path('one_market_data/one_tick')
    res_path = None

    paths = []
    pythonpath = os.environ.get('PYTHONPATH')
    if pythonpath:
        paths = pythonpath.split(os.pathsep)
    main_one_tick_dir = os.environ.get('MAIN_ONE_TICK_DIR')
    if main_one_tick_dir:
        paths.append(main_one_tick_dir)

    for value in paths:
        if str(omd_prefix) in str(Path(value)):
            res_path = value
            break

    if res_path is None:
        raise FileNotFoundError('Path to OneTick distribution is not found')

    base_prefix = 'one_market_data'
    omd_prefix_pos = res_path.find(base_prefix)
    res_path = res_path[: omd_prefix_pos + len(base_prefix)]

    return res_path



from locator_parser.common import Entity
from locator_parser.actions import DoNothing


class Allow(Entity):
    TAG = "ALLOW"
    SINGLE = True


class DB(Entity):
    TAG = "DB"
    HAS_PROPERTIES = True
    CHILDREN = [Allow]


class DBs(Entity):
    TAG = "DATABASES"
    CHILDREN = [DB]


class User(Entity):
    TAG = "USER"
    SINGLE = True


class Role(Entity):
    TAG = "ROLE"
    HAS_PROPERTIES = True
    CHILDREN = [User]


class Roles(Entity):
    TAG = "ROLES"
    CHILDREN = [Role]


class EP(Entity):
    TAG = "EP"
    HAS_PROPERTIES = True
    CHILDREN = [Allow]


class EPs(Entity):
    TAG = "EVENT_PROCESSORS"
    CHILDREN = [EP]


def parse_acl(reader, writer, action=DoNothing()):
    writer.refresh()
    reader.set_writer(writer)

    roles_p = Roles()
    databases_p = DBs()
    eps_p = EPs()

    for _ in reader:
        roles_p(reader, writer, action)
        databases_p(reader, writer, action)
        eps_p(reader, writer, action)


# ----------------------------------------- #
# set parent
for _, cls in list(globals().items()):
    try:
        if issubclass(cls, Entity) and cls != Entity:
            for child in cls.CHILDREN:
                child.PARENT = Entity

                if not cls.HAS_PROPERTIES and not cls.SINGLE:
                    child.PARENT = cls
    except Exception:
        continue



import types
from typing import Union

from collections import defaultdict

from ._internal._state_vars import StateVars
from ._internal._state_objects import _StateBase, check_field_name_in_schema
from .column import _Column
from .column_operations.base import _Operation
from .. import types as ott
from ._source.symbol import SymbolType


class _CompareTrackScope:
    """
    We need it to prevent problems when some exceptions are risen, and we left non-default tracking settings.
    This scope guarantee that code works always with the same state of tracker.
    """
    def __init__(self, emulation_enabled: bool = True):
        self.emulation_enabled = emulation_enabled

    def __enter__(self):
        _Operation.emulation_enabled = self.emulation_enabled
        _Column.emulation_enabled = self.emulation_enabled

    def __exit__(self, exc_type, exc_val, exc_tb):
        _Operation.emulation_enabled = not self.emulation_enabled
        _Column.emulation_enabled = not self.emulation_enabled


class _EmulateStateVars(StateVars):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _EmulateStateVars.CHANGED_TICK_LISTS = {}

    def __setitem__(self, key, value):
        if key not in self._columns:
            raise ValueError("Can't declare state variables in per-tick script")

        if isinstance(value, _Operation):
            value = self._set_var_from_operation(key, value)
        elif isinstance(value, _StateBase):
            value = value.copy(obj_ref=self, name=key)

        return f'{str(self._columns[key])} = {ott.value2str(value)};'


class _EmulateInputObject:

    Symbol = SymbolType()  # NOSONAR

    def __init__(self, parent_obj):
        self.__class__._state_vars = _EmulateStateVars(self, parent_obj)

        for name, dtype in parent_obj.columns().items():
            self.__dict__[name] = _Column(name, dtype, self)
            if name == 'TIMESTAMP':
                self.__dict__['Time'] = _Column(name, dtype, self)

        # creating class variables, so they will not go to the self.__dict__
        # they will not be used before object initialization anyway

        # collects new columns with appeared values
        self.__class__.NEW_VALUES = defaultdict(lambda: [])
        self.__class__.LOCAL_VARS_NEW_VALUES = defaultdict(lambda: [])
        # collects local variables in the per-tick-script
        self.__class__.LOCAL_VARS = {}
        self.__class__.STATIC_VARS = {}
        # dictionary with functions used in per-tick script
        self.__class__.FUNCTIONS = {}
        self.__class__.INPUT_SCHEMA = self._get_schema()

    def _get_schema(self):
        return {
            name: value.dtype
            for name, value in self.__dict__.items()
            if isinstance(value, _Column)
        }

    @property
    def schema(self):
        return self.INPUT_SCHEMA

    def __getitem__(self, item):
        if item not in self.__dict__:
            raise NameError(f"Column '{item}' referenced before assignment")

        return self.__getattr__(item)

    def __getattr__(self, item):
        if item not in self.__dict__:
            raise AttributeError(f"There is no '{item}' attribute")

        item = self.__dict__[item]

        if not isinstance(item, _Column):
            return item

        dtype = item.dtype
        name = item.name
        if issubclass(dtype, ott.nsectime):
            return self.get_datetime_value(name)
        if issubclass(dtype, int):
            return self.get_long_value(name)
        if issubclass(dtype, float):
            return self.get_double_value(name)
        if issubclass(dtype, str):
            return self.get_string_value(name)
        raise AttributeError(f'{item} has {dtype} type, so it is not possible to return correct get method')

    @classmethod
    def get_changed_tick_lists(cls):
        return cls._state_vars.CHANGED_TICK_LISTS

    @property
    def state_vars(self):
        return self._state_vars

    def __str__(self):
        return 'LOCAL::INPUT_TICK'

    def copy_tick(self, tick_object):
        """
        Copy fields from ``tick_obj`` to the output tick.
        Will only rewrite fields that are presented in this tick and ``tick_object``,
        will not remove or add any.
        Translated to COPY_TICK() function.
        """
        return f'COPY_TICK({tick_object});'

    def get_long_value(self, field_name: Union[str, _Operation]) -> _Operation:
        """
        Get value of the long ``field_name`` of the tick.

        Parameters
        ----------
        field_name: str, :py:class:`otp.Operation <onetick.py.core.column_operations.base.Operation>`
            String field name or operation which returns field name.

        Examples
        --------
        >>> def fun(tick):
        ...     tick['TOTAL_INT'] = 0
        ...     for field in otp.tick_descriptor_fields():
        ...         if field.get_type() == 'long':
        ...             tick['TOTAL_INT'] += tick.get_long_value(field.get_name())
        >>> t = otp.Tick(INT_1=3, INT_2=5)
        >>> t = t.script(fun)
        >>> otp.run(t)
                Time  INT_1  INT_2  TOTAL_INT
        0 2003-12-01      3      5          8
        """
        schema = self.schema
        check_field_name_in_schema(field_name, int, schema=schema)
        return _Operation(dtype=int,
                          op_str=f'{self}.GET_LONG_VALUE({ott.value2str(field_name)})')

    def get_double_value(self, field_name: Union[str, _Operation]) -> _Operation:
        """
        Get value of the double ``field_name`` of the tick.

        Parameters
        ----------
        field_name: str, :py:class:`otp.Operation <onetick.py.core.column_operations.base.Operation>`
            String field name or operation which returns field name.

        Examples
        --------
        >>> def fun(tick):
        ...     tick['TOTAL_DOUBLE'] = 0.0
        ...     for field in otp.tick_descriptor_fields():
        ...         if field.get_type() == 'double':
        ...             tick['TOTAL_DOUBLE'] += tick.get_double_value(field.get_name())
        >>> t = otp.Tick(DOUBLE_1=3.1, DOUBLE_2=5.2)
        >>> t = t.script(fun)
        >>> otp.run(t)
                Time  DOUBLE_1  DOUBLE_2  TOTAL_DOUBLE
        0 2003-12-01      3.1      5.2          8.3
        """
        check_field_name_in_schema(field_name, float, schema=self.schema)
        return _Operation(dtype=float,
                          op_str=f'{self}.GET_DOUBLE_VALUE({ott.value2str(field_name)})')

    def get_string_value(self, field_name: Union[str, _Operation]) -> _Operation:
        """
        Get value of the string ``field_name`` of the tick.

        Parameters
        ----------
        field_name: str, :py:class:`otp.Operation <onetick.py.core.column_operations.base.Operation>`
            String field name or operation which returns field name.

        Examples
        --------
        >>> def fun(tick):
        ...     tick['TOTAL_STR'] = ""
        ...     for field in otp.tick_descriptor_fields():
        ...         if field.get_type() == 'string':
        ...             tick['TOTAL_STR'] += tick.get_string_value(field.get_name())
        >>> t = otp.Tick(STR_1="1", STR_2="2")
        >>> t = t.script(fun)
        >>> otp.run(t)
                Time  STR_1  STR_2  TOTAL_STR
        0 2003-12-01      1      2          12
        """
        check_field_name_in_schema(field_name, str, schema=self.schema)
        return _Operation(dtype=str,
                          op_str=f'{self}.GET_STRING_VALUE({ott.value2str(field_name)})')

    def get_datetime_value(self, field_name: Union[str, _Operation]) -> _Operation:
        """
        Get value of the datetime ``field_name`` of the tick.

        Parameters
        ----------
        field_name: str, :py:class:`otp.Operation <onetick.py.core.column_operations.base.Operation>`
            String field name or operation which returns field name.

        Examples
        --------
        >>> def fun(tick):
        ...     for field in otp.tick_descriptor_fields():
        ...         if field.get_type() == 'nsectime':
        ...             tick['SOME_DATETIME'] = tick.get_datetime_value(field.get_name())
        >>> t = otp.Tick(DATETIME=otp.datetime(2021, 1, 1))
        >>> t = t.script(fun)
        >>> otp.run(t)
                Time   DATETIME SOME_DATETIME
        0 2003-12-01 2021-01-01    2021-01-01
        """
        check_field_name_in_schema(field_name, ott.nsectime, schema=self.schema)
        return _Operation(dtype=ott.nsectime,  # we don't know if it is msectime
                          op_str=f'{self}.GET_DATETIME_VALUE({ott.value2str(field_name)})')


class _EmulateObject(_EmulateInputObject):
    """
    Instances of this class are proxy to columns of a _Source object,
    and track assignment to construct the correct per-tick-script then
    """
    def __init__(self, parent_obj):
        super().__init__(parent_obj)
        _EmulateObject.input = _EmulateInputObject(parent_obj)

    def __getattr__(self, item):
        if item not in self.__dict__:
            raise AttributeError(f"There is no '{item}' attribute")

        return self.__dict__[item]

    def __setitem__(self, key, value):
        return self.__setattr__(key, value)

    def __setattr__(self, key, value):
        if key in self.__class__.__dict__:
            super().__setattr__(key, value)
        if key not in self.__dict__ or key in _EmulateObject.NEW_VALUES:
            _EmulateObject.NEW_VALUES[key].append(value)
        if key not in self.__dict__:
            self.__dict__[key] = _Column(key, ott.get_object_type(value), self)

    def __str__(self):
        return 'LOCAL::OUTPUT_TICK'

    @property
    def schema(self):
        return self._get_schema()

    @classmethod
    def get_types_of_new_columns(cls):
        """ method calculates types for all tracked new columns """
        return {
            key: ott.get_type_by_objects(values)
            for key, values in _EmulateObject.NEW_VALUES.items()
        }

    def set_long_value(self, field_name: Union[str, _Operation], value: Union[int, _Operation]) -> str:
        """
        Set ``value`` of the long ``field_name`` of the tick.

        Parameters
        ----------
        field_name: str, :py:class:`otp.Operation <onetick.py.core.column_operations.base.Operation>`
            String field name or operation which returns field name.
        value: int, :py:class:`otp.Operation <onetick.py.core.column_operations.base.Operation>`
            Long value to set or operation which return such value.

        Examples
        --------
        >>> def fun(tick):
        ...     for field in otp.tick_descriptor_fields():
        ...         if field.get_type() == 'long':
        ...             tick.set_long_value(field.get_name(), 5)
        >>> t = otp.Tick(INT_1=3)
        >>> t = t.script(fun)
        >>> otp.run(t)
                Time  INT_1
        0 2003-12-01      5
        """
        check_field_name_in_schema(field_name, int, value, self.schema)
        return f'{self}.SET_LONG_VALUE({ott.value2str(field_name)}, {ott.value2str(value)});'

    def set_double_value(self, field_name: Union[str, _Operation], value: Union[float, _Operation]) -> str:
        """
        Set ``value`` of the double ``field_name`` of the tick.

        Parameters
        ----------
        field_name: str, :py:class:`otp.Operation <onetick.py.core.column_operations.base.Operation>`
            String field name or operation which returns field name.
        value: float, :py:class:`otp.Operation <onetick.py.core.column_operations.base.Operation>`
            Double value to set or operation which return such value.

        Examples
        --------
        >>> def fun(tick):
        ...     for field in otp.tick_descriptor_fields():
        ...         if field.get_type() == 'double':
        ...             tick.set_double_value(field.get_name(), 5.0)
        >>> t = otp.Tick(DOUBLE_1=3.0)
        >>> t = t.script(fun)
        >>> otp.run(t)
                Time  DOUBLE_1
        0 2003-12-01      5.0
        """
        check_field_name_in_schema(field_name, float, value, self.schema)
        return f'{self}.SET_DOUBLE_VALUE({ott.value2str(field_name)}, {ott.value2str(value)});'

    def set_string_value(self, field_name: Union[str, _Operation], value: Union[str, _Operation]) -> str:
        """
        Set ``value`` of the string ``field_name`` of the tick.

        Parameters
        ----------
        field_name: str, :py:class:`otp.Operation <onetick.py.core.column_operations.base.Operation>`
            String field name or operation which returns field name.
        value: str, :py:class:`otp.Operation <onetick.py.core.column_operations.base.Operation>`
            String value to set or operation which return such value.

        Examples
        --------
        >>> def fun(tick):
        ...     for field in otp.tick_descriptor_fields():
        ...         if field.get_type() == 'string':
        ...             tick.set_string_value(field.get_name(), '5')
        >>> t = otp.Tick(STR_1='3')
        >>> t = t.script(fun)
        >>> otp.run(t)
                Time  STR_1
        0 2003-12-01      5
        """
        check_field_name_in_schema(field_name, str, value, self.schema)
        return f'{self}.SET_STRING_VALUE({ott.value2str(field_name)}, {ott.value2str(value)});'

    def set_datetime_value(self, field_name: Union[str, _Operation], value: Union[int, _Operation]) -> str:
        """
        Set ``value`` of the datetime ``field_name`` of the tick.

        Parameters
        ----------
        field_name: str, :py:class:`otp.Operation <onetick.py.core.column_operations.base.Operation>`
            String field name or operation which returns field name.
        value: int, :py:class:`otp.Operation <onetick.py.core.column_operations.base.Operation>`
            Datetime value to set or operation which return such value.

        Examples
        --------
        >>> def fun(tick):
        ...     for field in otp.tick_descriptor_fields():
        ...         if field.get_type() == 'nsectime':
        ...             tick.set_datetime_value(field.get_name(), otp.datetime(2021, 1, 1) - otp.Day(1))
        >>> t = otp.Tick(DATETIME_1=otp.datetime(2021, 1, 1))
        >>> t = t.script(fun)
        >>> otp.run(t)
                Time  DATETIME_1
        0 2003-12-01      2020-12-31
        """
        check_field_name_in_schema(field_name, ott.nsectime, value, self.schema)
        return f'{self}.SET_DATETIME_VALUE({ott.value2str(field_name)}, {ott.value2str(value)});'


def _validate_lambda(lambda_f):
    if not (isinstance(lambda_f, types.LambdaType) and lambda_f.__name__ == '<lambda>'
            or isinstance(lambda_f, (types.FunctionType, types.MethodType))):
        raise ValueError("It is expected to get a function, method or lambda,"
                         f" but got '{lambda_f}' of type '{type(lambda_f)}'")


def apply_script(lambda_f, self_ref):
    """
    In this function we parse python syntax tree for `lambda_f`
    and converting it to OneTick's per-tick script.
    This function returns _Column like function, that
    keeps resulting built expression.
    """
    _validate_lambda(lambda_f)

    with _CompareTrackScope():
        # TODO: remove circular imports
        from .per_tick_script import FunctionParser
        _script = FunctionParser(lambda_f, emulator=self_ref).per_tick_script()
        new_columns_types = _EmulateObject.get_types_of_new_columns()
        return new_columns_types, _script


def apply_lambda(lambda_f, self_ref):
    """
    In this function we parse python syntax tree for `lambda_f`
    and converting it to OneTick's CASE function().
    This function returns _Column like function, that
    keeps resulting built expression.
    """
    _validate_lambda(lambda_f)

    with _CompareTrackScope():
        # TODO: remove circular imports
        from .per_tick_script import FunctionParser
        res, values = FunctionParser(lambda_f, emulator=self_ref).case()
        return _LambdaIfElse(res, ott.get_type_by_objects(values))


class _LambdaIfElse(_Column):
    """
    Behaves like a column and consists information built from the lambda calculation
    """

    def __init__(self, repr, dtype):
        self._dtype = dtype
        self._repr = repr

    def __str__(self):
        return self._repr

    def __len__(self):
        if issubclass(self.dtype, str):
            if self.dtype is str:
                return ott.string.DEFAULT_LENGTH
            else:
                return self.dtype.length

        raise TypeError(f"It is not allowed to call len() method for object of type '{self.dtype}'")



import os
import re
import io

from collections import defaultdict

from onetick.py.utils import abspath_to_query_by_name

MULTIPLE_NAMESAKE_PARAMETERS_ALLOWED = frozenset(
    [
        "BIND_SECURITY",
        # "PARAMETER" isn't on the list as it's handled separately
    ]
)


class QueryNotFoundError(FileNotFoundError):
    pass


class UncertainQueryName(ValueError):
    pass


def get_queries(otq_path):
    result = []
    with open(otq_path, "r") as fin:
        for line in fin:
            line = line.strip()

            res = re.match(r"^\[(.*)\]$", line)
            if res:
                name = res.group(1)
                if name != "_meta":
                    result.append(name)

    return result


def is_commented_out(node):
    return hasattr(node, "COMMENTED_OUT") and (node.COMMENTED_OUT == "1")


def get_query_info(otq_path, query_name="", inspected_graphs=None):
    """
    Collects some meta information about an .otq-based query.

    Parameters
    ----------
    otq_path: str
        Absolute path to an OTQ file.
    query_name: str
        Name of the query to analyze; if it's not specified, the .otq file should have only one query.
    inspected_graphs: None or Dict[tuple of (str, str) : Graph]
        Usually should be kept as the default value (None).

    Returns
    -------

    Graph - Instance of an internal class. Contains the following fields:
        nodes: list of Node
            Nodes in the query.
        sources: list of Node
            Nodes without incoming links.
        sinks: list of Node
            Nodes without outcoming links.
        nested_inputs: list of Node
            Nodes with nested inputs.
        nested_outputs: list of Node
            Nodes with nested outputs.
        has_unbound_sources: bool
            Whether the query needs an unbound security list to be provided.
    """
    if inspected_graphs is None:
        inspected_graphs = {}

    if (otq_path, query_name) in inspected_graphs:
        return inspected_graphs[(otq_path, query_name)]

    parsed = False
    found = False

    class Node:
        def __init__(self):
            self.SINK = []
            self.SOURCE = []

        def _get_ep_name_and_stack_info(self):
            if not hasattr(self, 'EP'):
                return None, None
            # self.EP can be multiline, so using re.DOTALL for dot to match newline too
            res = re.match(r"([A-Z_]+?)(\((.+)\))?$", self.EP, flags=re.DOTALL)
            if not res:
                return None, None
            ep_name = res.group(1)
            ep_parameters = res.group(3)
            stack_info = None
            if ep_parameters and 'STACK_INFO' in ep_parameters:
                res = re.match(r'.*STACK_INFO="?(-?\d+)', ep_parameters)
                if res:
                    stack_info = res.group(1)
            return ep_name, stack_info

    graph = defaultdict(Node)

    if not os.path.exists(otq_path):
        raise FileNotFoundError(f'otq "{otq_path}" is not found')

    if not otq_path.endswith('.py'):

        multiline_ep = False
        multiline_value = ''

        with open(otq_path, "r") as fin:
            for line in fin:
                line = line.strip()

                if line.startswith("[") and line.endswith("]"):
                    if ("[" + query_name + "]") == line:
                        found = True
                        continue
                    else:
                        if line != "[_meta]" and query_name == "":
                            if found is True:
                                raise UncertainQueryName(
                                    "No query name was passed, implying the file has only one query; it has more"
                                )
                            else:
                                found = True
                                continue

                if (found and not parsed) or multiline_ep:
                    if line.startswith("TYPE = "):
                        parsed = True
                        continue

                    if not multiline_ep:
                        res = re.match(r"NODE_?(\d+)_?(.*?)\s?=\s?(.*)", line)  # NOSONAR

                        if res:
                            num = int(res.group(1))
                            param = res.group(2)
                            value = res.group(3)
                        else:
                            res = re.match(r"ROOT_?(.*?)\s?=\s?(.*)", line)

                            if res:
                                num = 0
                                param = res.group(1)
                                value = res.group(2)

                        if res and line.endswith('\\'):
                            # first line of multiline
                            multiline_ep = True
                            multiline_value += value.replace('\\', '\n')
                            continue
                    else:
                        # next line of multiline
                        multiline_value += line.replace('\\', '\n')
                        if line.endswith('\\'):
                            continue
                        # last line of multiline
                        value = multiline_value
                        multiline_ep = False
                        multiline_value = ''

                    if res:
                        param, value = param.strip(), value.strip()

                        if param == "PARAMETER":
                            if not hasattr(graph[num], "PARAMETERS"):
                                setattr(graph[num], "PARAMETERS", {})
                            bits = value.split(" ")
                            inner_param_name = bits[0]
                            inner_param_value = " ".join(bits[1:])
                            graph[num].PARAMETERS[inner_param_name] = inner_param_value
                            continue

                        if param in ("SOURCE", "SINK"):
                            sources = []

                            for v in value.split():
                                if v == "ROOT":
                                    sources.append(0)
                                else:
                                    res = re.match(r"NODE_?(\d+)(\.(.*))*", v)
                                    sources.append(int(res.group(1)))

                            value = sources
                        elif param == "":
                            param = "EP"

                        if param in MULTIPLE_NAMESAKE_PARAMETERS_ALLOWED:
                            # A node can have several parameters with the same name
                            if not hasattr(graph[num], param):
                                setattr(graph[num], param, [])

                            getattr(graph[num], param).append(value)
                        else:
                            setattr(graph[num], param, value)

        if not found:
            raise QueryNotFoundError(f'Query "{query_name}" is not found in the {otq_path}')

    for num, node in graph.items():
        setattr(node, "NUM", num)
        setattr(node, "IS_NESTED", False)

        if hasattr(node, "NESTED_INPUT"):
            node.NESTED_INPUT = node.NESTED_INPUT.split()[0]

        if hasattr(node, "NESTED_OUTPUT"):
            node.NESTED_OUTPUT = node.NESTED_OUTPUT.split()[0]

        for src in node.SOURCE:
            if num not in graph[src].SINK:
                graph[src].SINK.append(num)

        for src in node.SINK:
            if num not in graph[src].SOURCE:
                graph[src].SOURCE.append(num)

        if not is_commented_out(node):
            # Commented nodes can affect sources and sinks, but cannot anything else
            if node.EP.startswith("NESTED_OTQ"):
                setattr(node, "IS_NESTED", True)
                if node.EP.strip() == "NESTED_OTQ":
                    if "OTQ_PATH" in node.PARAMETERS:
                        address = node.PARAMETERS["OTQ_PATH"]
                    elif "OTQ_NODE" in node.PARAMETERS:
                        address = node.PARAMETERS["OTQ_NODE"]
                    elif hasattr(node, "PH_PATH"):
                        address = node.PH_PATH
                    else:
                        raise ValueError("Can't get NESTED_OTQ address")

                    node.EP = "NESTED_OTQ " + address

                setattr(node, "NESTED_GRAPH", _load_nested_query(otq_path, node.EP, inspected_graphs))

    class Graph:
        def __init__(self, nodes):
            self.nodes = nodes
            self.sources = [node for _, node in graph.items() if len(node.SOURCE) == 0]
            self.sinks = [node for _, node in graph.items() if len(node.SINK) == 0]

            self.nested_inputs = [node for node in self.sources if hasattr(node, "NESTED_INPUT")]
            self.nested_outputs = [node for node in self.sinks if hasattr(node, "NESTED_OUTPUT")]

            inspected_graphs[(otq_path, query_name)] = self

            self.has_unbound_sources = self._check_for_unbound_sources()

        def _check_for_unbound_sources(self, bound_symbol_info=None):
            if bound_symbol_info is None:
                bound_symbol_info = {}

            for node in self.sources:
                self._search_for_bound_sink(node.NUM, bound_symbol_info)

            has_unbound_sources = any((not bound_symbol_info[node.NUM]) for node in self.sources)
            return has_unbound_sources

        def _search_for_bound_sink(self, root, bound_symbol_info):
            if root in bound_symbol_info:
                return

            if not is_commented_out(self.nodes[root]):
                # A commented node cannot have its own bound symbols, but can have a bound sink
                if hasattr(self.nodes[root], "BIND_SECURITY"):
                    for bound_security in self.nodes[root].BIND_SECURITY:
                        if (
                            ("_SYMBOL_NAME" not in bound_security)
                            and ("_SYMBOL_PARAM" not in bound_security)
                            and (not bound_security.endswith("No"))
                        ):
                            # The first two define dependency on lower-bound symbol;
                            # The last is whether the bound security is not unchecked in the list

                            bound_symbol_info[root] = True
                            break

                if self.nodes[root].IS_NESTED:
                    nested = self.nodes[root].NESTED_GRAPH
                    if not nested.has_unbound_sources:
                        bound_symbol_info[root] = True

            if root not in bound_symbol_info:
                for node in self.nodes[root].SINK:
                    if node not in bound_symbol_info:
                        self._search_for_bound_sink(node, bound_symbol_info)

                    if bound_symbol_info.get(node, None) is True:
                        bound_symbol_info[root] = True
                        break

                if root not in bound_symbol_info:
                    bound_symbol_info[root] = False

        def has_unbound_if_pinned(self, in_pins):
            """
            Parameters
            ----------
            in_pins: dict(str: bool)
                mapping of pins to whether the respective source needs unbound symbols

            Returns
            -------
            True if query with pins bound to graphs with described boundness state would need unbound symbols.
            """

            if not self.has_unbound_sources:
                # All sources are bound inside the query graph.
                return False

            in_pins_to_nodes = {node.NESTED_INPUT: node for node in self.nested_inputs}

            bound_symbol_info = {}

            for pin, needs_unbound in in_pins.items():
                if not needs_unbound:
                    bound_symbol_info[in_pins_to_nodes[pin].NUM] = True

            return self._check_for_unbound_sources(bound_symbol_info)

    return Graph(graph)


def _load_nested_query(nesting_file, nested_address, inspected_graphs):
    nested_address = nested_address.split(" ")[1]
    address_bits = nested_address.split("::")

    if len(address_bits) == 1:
        # Only the path is given, no query name
        # E.g. utils/FindSymbols.otq
        query_file = address_bits[0]
        query = ""
    elif address_bits[-1].endswith(".otq"):
        # remote://DB::utils/FindSymbols.otq
        query_file = address_bits[-1]
        query = ""
    else:
        # remote://DB::utils/FindSymbols.otq::Find
        query_file, query = address_bits[-2:]

    if query_file == "___ME___":
        otq_path = nesting_file
    else:
        otq_path = abspath_to_query_by_name(query_file)

    nested_graph = get_query_info(otq_path, query, inspected_graphs)
    return nested_graph


def add_pins(otq_path, query_name, specification):
    """
    The function takes a query and adds pins to the query according to the
    specification, and saves it back to the original file.

    Parameters
    ----------
    otq_path: str
        Absolute path to an OTQ file.
    query_name: str
        Name of the query to analyze
    specification: List[tuple]
        List of 3-values tuples, where the first value node from the `get_query_info` function,
        the second one is marker, the third one is name.
        A marker values: 1 for input, and 0 for output.

    Returns
    -------
    Nothing

    Raises
    ------
    FileNotFoundError
    """
    if not os.path.exists(otq_path):
        raise FileNotFoundError(f'otq "{otq_path}" is not found')

    found = False
    out = io.StringIO()

    with open(otq_path, "r") as fin:

        multiline_ep = False

        for line in fin:
            line = line.strip()

            if line.startswith("[") and line.endswith("]"):
                if ("[" + query_name + "]") == line:
                    found = True

            out.write(line + "\n")

            if found or multiline_ep:
                if not multiline_ep:
                    res = re.match(r"NODE_?(\d+)_?(.*?)\s?=\s?(.*)", line)  # NOSONAR

                    if res:
                        num = int(res.group(1))
                    else:
                        res = re.match(r"ROOT_?(.*?)\s?=\s?(.*)", line)

                        if res:
                            num = 0

                    if res and line.endswith('\\'):
                        # first line of multiline
                        multiline_ep = True
                        continue
                else:
                    if line.endswith('\\'):
                        continue
                    multiline_ep = False

                if res:
                    for inx, v in enumerate(specification):
                        node, pin_flag, pin_name = v
                        if node.NUM == num:  # noqa
                            if pin_flag is None:
                                continue
                            node_name = "NODE_" + str(num)
                            if num == 0:
                                node_name = "ROOT"

                            pin_id = "NESTED_INPUT" if pin_flag == 1 else "NESTED_OUTPUT"

                            out.write(node_name + "_" + pin_id + " = " + pin_name + "\n")

                            # don't use this specification item anymore
                            specification[inx] = (node, None, None)

    with open(otq_path, "w") as fout:
        fout.write(out.getvalue())


def get_query_parameter_list(otq_path, query_name):
    """Returns a list of query parameter names; can be used for nesting to pass the parameters to the nested query"""
    if not os.path.exists(otq_path):
        raise FileNotFoundError(f'otq "{otq_path}" is not found')

    found = False
    param_list = []

    with open(otq_path, "r") as fin:
        for line in fin:
            line = line.strip()

            if line.startswith("[") and line.endswith("]"):
                if ("[" + query_name + "]") == line:
                    found = True
                else:
                    found = False

            if found:
                params = re.findall(r'\$(\w+|(\{(\w+)\}))', line)
                if params:
                    param_list.extend([p[2] if p[2] != '' else p[0] for p in params])

    param_list.sort()
    return param_list



from onetick.py.otq import otq


class CallbackBase(otq.CallbackBase):
    """
    Base class for user-defined callback classes
    for :py:func:`onetick.py.run` callback mode.

    Note
    ----
    Callbacks are executed sequentially, so make sure
    they don't take too much time to execute.

    See also
    --------
    :py:func:`onetick.py.run`

    Examples
    --------
    >>> t = otp.Ticks(A=[1, 2, 3])
    >>> class NumTicksCallback(otp.CallbackBase):
    ...     def __init__(self):
    ...         self.num_ticks = 0
    ...     def process_tick(self, tick, time):
    ...         self.num_ticks += 1
    >>> callback = NumTicksCallback()
    >>> otp.run(t, callback=callback)
    >>> callback.num_ticks
    3
    """

    # pylint: disable-next=useless-parent-delegation
    def __init__(self):
        """
        Method ``__init__()`` can be used for callback initialization.
        Can be used to define some variables for future use
        in callback methods.
        """
        super().__init__()

    def replicate(self):
        """
        Called to replicate the callback object for each output node.
        May also be used for internal copying of callback object.

        Returns
        -------
            By default reference to this callback object
        """
        return self

    def process_callback_label(self, callback_label):
        """
        Called immediately before :meth:`process_symbol_name`
        to supply label assigned to callback.

        Parameters
        ----------
        callback_label: str
            label assigned to this callback object
        """
        pass

    def process_symbol_name(self, symbol_name):
        """
        Invoked to supply the name of the security that produces
        all the ticks that will be delivered to this callback object.
        If these ticks are provided by several securities,
        the ``symbol_name`` parameter is set to empty string.

        Parameters
        ----------
        symbol_name: str
            name of security

        Examples
        --------
        >>> t = otp.Tick(A=1)
        >>> class SymbolNameCallback(otp.CallbackBase):
        ...     def process_symbol_name(self, symbol_name):
        ...         self.symbol_name = symbol_name
        >>> callback = SymbolNameCallback()
        >>> otp.run(t, callback=callback, symbols='DEMO_L1::X')
        >>> callback.symbol_name
        'DEMO_L1::X'
        """
        pass

    def process_symbol_group_name(self, symbol_group_name):
        """
        Called when a named group of securities, i.e. portfolio, is processed.

        Parameters
        ----------
        symbol_group_name: str
            The name of security group.
        """
        pass

    def process_tick_type(self, tick_type):
        """
        Reports the type of the security ticks which are processed by this callback object.
        This method is called before any call to
        :meth:`process_tick_descriptor` or :meth:`process_tick`.
        It is called immediately after :meth:`process_symbol_name`.

        Parameters
        ----------
        tick_type: str
            The name of tick type.

        Examples
        --------
        >>> t = otp.Tick(A=1, symbol='DEMO_L1', tick_type='TT_TT')
        >>> class TickTypeCallback(otp.CallbackBase):
        ...     def process_tick_type(self, tick_type):
        ...         self.tick_type = tick_type
        >>> callback = TickTypeCallback()
        >>> otp.run(t, callback=callback)
        >>> callback.tick_type
        'DEMO_L1::TT_TT'
        """
        pass

    def process_tick_descriptor(self, tick_descriptor):
        """
        This method is invoked before the first call to :meth:`process_tick`
        and every time before tick structure changes.

        Parameters
        ----------
        tick_descriptor: list of tuple
            First element of each tuple is field's name
            and the second one is a dictionary ``{'type': string_field_type}``.

        Examples
        --------
        >>> t = otp.Tick(A=1)
        >>> class TickDescriptorCallback(otp.CallbackBase):
        ...    def process_tick_descriptor(self, tick_descriptor):
        ...        self.tick_descriptor = tick_descriptor
        >>> callback = TickDescriptorCallback()
        >>> otp.run(t, callback=callback)
        >>> callback.tick_descriptor
        [('A', {'type': 'int'})]
        """
        pass

    def process_tick(self, tick, time):
        """
        Called to deliver each tick.

        Note
        ----
        If you are making query through WebAPI mode, use ``process_ticks`` callback method instead.

        Parameters
        ----------
        tick: dict
            mapping of field names to field values
        time: :py:class:`datetime.datetime`
            timestamp of the tick in GMT timezone.

        Examples
        --------
        >>> t = otp.Tick(A=1)
        >>> class ProcessTickCallback(otp.CallbackBase):
        ...     def process_tick(self, tick, time):
        ...         self.result = (tick, time)
        >>> callback = ProcessTickCallback()
        >>> otp.run(t, callback=callback)
        >>> callback.result
        ({'A': 1}, datetime.datetime(2003, 12, 1, 5, 0))
        """
        pass

    def process_ticks(self, ticks):
        """
        Called after getting all ticks for WebAPI queries.

        Due to limitation of WebAPI mode ``process_tick`` callback method, which invoked on each tick, isn't supported.
        Instead, WebAPI supports ``process_ticks`` method which invoked for processing the data
        after the query is finished.
        All ticks are returned on the ``ticks`` variable.

        Parameters
        ----------
        ticks: dict
            mapping of field names to field values

        Examples
        --------
        >>> t = otp.Tick(A=1)
        >>> class ProcessTicksCallback(otp.CallbackBase):
        ...     def __init__(self):
        ...         self.result = {}
        ...
        ...     def process_ticks(self, ticks):
        ...         self.result = ticks
        >>> callback = ProcessTicksCallback()
        >>> otp.run(t, callback=callback)  # doctest: +SKIP
        >>> callback.result  # doctest: +SKIP
        {'Time': array(['2003-12-01T00:00:00.000000000'], dtype='datetime64[ns]'), 'A': array([1])}
        """
        pass

    def process_sorting_order(self, sorted_by_time_flag):
        """
        Informs whether the ticks that will be submitted to this callback object
        will be ordered by time.

        Parameters
        ----------
        sorted_by_time_flag: bool
            Indicates whether the incoming ticks will be sorted by time.
        """
        pass

    def process_data_quality_change(self, symbol_name, data_quality, time):
        """
        Called to report a data quality change, such as collection outage.

        Parameters
        ----------
        symbol_name: str
            Symbol name for each data quality change is propagated.
        data_quality: int
            parameter has following meaning:

            * `QUALITY_UNKNOWN` = -1,
            * `QUALITY_OK`,
            * `QUALITY_STALE` = 1,
            * `QUALITY_MISSING` = 2,
            * `QUALITY_PATCHED` = 4,
            * `QUALITY_MOUNT_BAD` = 9,
            * `QUALITY_DISCONNECT` = 17,
            * `QUALITY_COLLECTOR_FAILURE` = 33,
            * `QUALITY_DELAY_STITCHING_WITH_RT` = 64,
            * `QUALITY_OK_STITCHING_WITH_RT` = 66
        time: :py:class:`datetime.datetime`
            Time of the change in GMT timezone.
        """
        pass

    def process_error(self, error_code, error_msg):
        """
        Called to report a per-security error or per-security warning.

        Parameters
        ----------
        error_code: int
            Values of error code less than 1000 are warnings.
            Warnings signal issues which might not affect results of the query
            and thus could be chosen to be ignored
        error_msg: str
            Error message
        """
        pass

    def done(self):
        """
        Invoked when all the raw or computed ticks for a given request
        were submitted to the callback using the :meth:`process_tick` method.

        Examples
        --------
        >>> t = otp.Tick(A=1)
        >>> class DoneCallback(otp.CallbackBase):
        ...     def done(self):
        ...         self.done = True
        >>> callback = DoneCallback()
        >>> otp.run(t, callback=callback)
        >>> callback.done
        True
        """
        pass



import warnings

from onetick.py import types as ott
from onetick.py.types import value2str
from ._internal import MethodResult, _wrap_object, _type_error_for_op, _init_binary_op
from .op_types import (
    are_numerics, are_strings, are_bools, are_time, are_ints_not_time, _get_widest_type,
    are_floats, are_ints_or_time
)


class DatetimeSubtractionWarning(FutureWarning):
    pass


def round(prev_op, precision):
    if precision is not None:
        return MethodResult(f'round_double({str(prev_op)},{str(precision)})', float)
    else:
        return MethodResult(f'round({str(prev_op)})', int)


def isin(prev_op, items):
    if not items:
        raise ValueError("Method isin() can't be used without values")
    op_str = f'IN({str(prev_op)}, {",".join(value2str(i) for i in items)})'
    return MethodResult(op_str, bool)


def _map(prev_op, items, values_type, default=None):
    default_str = '' if default is None else f', {value2str(default)}'
    op_str = (f'CASE({str(prev_op)}, {",".join(value2str(k) + "," + value2str(v) for k, v in items.items())}'
              f'{default_str})')
    return MethodResult(op_str, values_type)


def fillna(prev_op, value=0):
    result = str(prev_op)

    if not issubclass(prev_op.dtype, float):
        raise TypeError(f"It is allowed to apply .fillna() is only columns that have 'float' type, "
                        f"but it is applied on a column with the {prev_op.dtype.__name__} type")
    dtype = ott.get_object_type(value)
    if not are_numerics(dtype):
        raise TypeError(f"fillna expects numeric type, but got value of type '{type(value)}'")
    op_str = f'replace_nan({result},{value2str(value)})'
    return MethodResult(op_str, float)


def add(prev_op, other):
    left, right, left_t, right_t, _, _ = _init_binary_op(prev_op, other)
    result = _return_dateadd_command(prev_op, other, left, right, left_t, right_t, "")
    # if it is DATEADD command return it, else try to form + operation
    if result:
        return result
    return _plus(prev_op, other, left, right, left_t, right_t)


def sub(prev_op, other):
    left, right, left_t, right_t, _, _ = _init_binary_op(prev_op, other)
    result = _return_dateadd_command(prev_op, other, left, right, left_t, right_t, "-")
    # if it is DATEADD command return it, else try to form - operation
    if result:
        return result
    return _minus(prev_op, other, left, right, left_t, right_t)


def _return_dateadd_command(prev_op, other, left, right, left_t, right_t, op_sign):
    if issubclass(left_t, ott.OTPBaseTimeOffset) and are_time(right_t):
        return _form_method_result_for_dateadd(right, right_t, op_sign, prev_op)
    elif issubclass(right_t, ott.OTPBaseTimeOffset) and are_time(left_t):
        return _form_method_result_for_dateadd(left, left_t, op_sign, other)
    return None


def _form_method_result_for_dateadd(op_str, dtype, op_sign, datepart):
    op_str = f"DATEADD({datepart.datepart}, {op_sign}({str(datepart.n)}), {op_str}, _TIMEZONE)"
    return MethodResult(op_str, dtype)


def _plus(prev_op, other, left, right, left_t, right_t):
    def get_str_len(op_or_str_const, dtype):
        if not isinstance(op_or_str_const, str):
            if dtype is str:
                return ott.string.DEFAULT_LENGTH
            else:
                return dtype.length
        elif isinstance(op_or_str_const, ott.varstring):
            return Ellipsis
        else:
            return len(op_or_str_const)  # ott.string as well as builtin strings have len operand

    op_str = f"{left} + {right}"
    if are_strings(left_t, right_t):  # strings support concatenation
        left_length, right_length = get_str_len(prev_op, left_t), get_str_len(other, right_t)
        if left_length is Ellipsis or right_length is Ellipsis:
            dtype = ott.varstring
        else:
            length = max(left_length, right_length)
            dtype = str if length <= ott.string.DEFAULT_LENGTH else ott.string[length]
    else:
        dtype = _get_dtype_for_plus_or_minus(left_t, right_t)
    if dtype:
        return MethodResult(op_str, dtype)
    else:
        raise _type_error_for_op("+", f"'{left_t}' and '{right_t}'")


def _minus(prev_op, other, left, right, left_t, right_t):  # noqa # NOSONAR
    op_str = f"{left} - {right}"
    # Between datetime (msectime and nsectime) types, only the - operator is allowed.
    if are_time(left_t, right_t):
        warnings.warn("Subtracting datetimes without specifying resulted time unit is deprecated. "
                      "Specify resulted time unit explicitly, ex: otp.Hour(a - b) or otp.Nano(a - b). "
                      "Difference in milliseconds will be returned now, but in future it will throw error. ",
                      DatetimeSubtractionWarning)
        dtype = int
        # TODO uncomment if we will decide to use nanosecond as default time unit
        # op_str = f"DATEDIFF('nanosecond', {right}, {left}, _TIMEZONE)"
        # return MethodResult(op_str, dtype)
    else:
        dtype = _get_dtype_for_plus_or_minus(left_t, right_t)
    if dtype:
        return MethodResult(op_str, dtype)
    else:
        raise _type_error_for_op("-", f"'{left_t}' and '{right_t}'")


def _get_dtype_for_plus_or_minus(left_t, right_t):
    dtype = None
    if are_numerics(left_t, right_t):
        dtype = _get_widest_type(left_t, right_t)
    # It is possible to add an integral value to datetime or subtract an integer from it,
    # in this case the type is not changed
    elif (are_ints_not_time(left_t) and issubclass(right_t, ott.nsectime)
          or are_ints_not_time(right_t) and issubclass(left_t, ott.nsectime)):
        dtype = ott.nsectime
    elif (are_ints_not_time(left_t) and issubclass(right_t, ott.msectime)
          or are_ints_not_time(right_t) and issubclass(left_t, ott.msectime)):
        dtype = ott.msectime
    # Any operation between floating-point and datetime types are not allowed in tick script,
    # but supported in EP, so we also allow such operation, but generate warning.
    elif (are_floats(left_t) and issubclass(right_t, ott.nsectime)
          or are_floats(right_t) and issubclass(left_t, ott.nsectime)):
        dtype = ott.nsectime
        warnings.warn("Onetick will shrink the fractional part")
    elif (are_floats(left_t) and issubclass(right_t, ott.msectime)
          or are_floats(right_t) and issubclass(left_t, ott.msectime)):
        dtype = ott.msectime
        warnings.warn("Onetick will shrink the fractional part")
    return dtype


def mul(prev_op, other):
    left, right, left_t, right_t, op_str, dtype = _init_binary_op(prev_op, other)

    if are_numerics(left_t, right_t):
        dtype = _get_widest_type(left_t, right_t)
        op_str = f"{left} * {right}"
    elif issubclass(left_t, str) and are_ints_not_time(right_t):
        op_str = f"repeat({left}, {right})"
        dtype = left_t
    elif issubclass(right_t, str) and are_ints_not_time(left_t):
        op_str = f"repeat({right}, {left})"
        dtype = right_t
    if dtype and op_str:
        return MethodResult(op_str, dtype)
    else:
        raise _type_error_for_op("*", f"'{left_t}' and '{right_t}'")


def div(prev_op, other):
    left, right, left_t, right_t, op_str, dtype = _init_binary_op(prev_op, other)
    if not are_numerics(left_t, right_t):
        raise _type_error_for_op("/", f"'{left_t}' and '{right_t}'")
    dtype = _get_widest_type(_get_widest_type(left_t, right_t), float)
    op_str = f"{left} / {right}"
    return MethodResult(op_str, dtype)


def mod(prev_op, other):
    left, right, left_t, right_t, op_str, dtype = _init_binary_op(prev_op, other)
    if not are_ints_not_time(left_t, right_t):
        raise _type_error_for_op("mod", f"'{left_t}' and '{right_t}'")
    dtype = int
    op_str = f"mod({left}, {right})"
    return MethodResult(op_str, dtype)


def abs(prev_op):
    dtype = ott.get_object_type(prev_op)
    if not are_numerics(dtype):
        raise TypeError(f"Operation is not supported for type '{dtype}'")
    op_str = f"abs{_wrap_object(prev_op)}"
    return MethodResult(op_str, dtype)


def pos(prev_op):
    dtype = ott.get_object_type(prev_op)
    if not are_numerics(dtype):
        raise TypeError(f"Operation is not supported for type '{dtype}'")
    op_str = f"(+{_wrap_object(prev_op)})"
    return MethodResult(op_str, dtype)


def neg(prev_op):
    dtype = ott.get_object_type(prev_op)
    if not are_numerics(dtype):
        raise TypeError(f"Operation is not supported for type '{dtype}'")
    op_str = f"(-{_wrap_object(prev_op)})"
    return MethodResult(op_str, dtype)


def invert(prev_op):
    dtype = ott.get_object_type(prev_op)
    if not issubclass(dtype, bool):
        raise TypeError(f"Operation is not supported for type '{dtype}'")
    op_str = f"not({_wrap_object(prev_op)})"
    return MethodResult(op_str, dtype)


def eq(prev_op, other):
    return _compare(prev_op, other, "=", "==")


def ne(prev_op, other):
    return _compare(prev_op, other, "!=")


def lt(prev_op, other):
    return _compare(prev_op, other, "<")


def le(prev_op, other):
    return _compare(prev_op, other, "<=")


def gt(prev_op, other):
    return _compare(prev_op, other, ">")


def ge(prev_op, other):
    return _compare(prev_op, other, ">=")


def _compare(prev_op, other, op_sign, op_sign_python_level=None):
    def one_boolean_operation_is_allowed(prev_op, other, left_t, right_t):
        return (isinstance(prev_op, bool) and issubclass(left_t, bool) or
                isinstance(other, bool) and issubclass(right_t, bool))

    def the_same_type(left_t, right_t):
        # Operands of compare operator are expected to be both numeric, both booleans, both matrices or both strings
        return (
            not are_bools(left_t) and not are_bools(right_t)
            and (
                are_numerics(left_t, right_t) or are_strings(left_t, right_t)
                or are_ints_or_time(left_t, right_t)
            )
            or are_bools(left_t, right_t)
        )

    left, right, left_t, right_t, op_str, _ = _init_binary_op(prev_op, other)
    if one_boolean_operation_is_allowed(prev_op, other, left_t, right_t) or the_same_type(left_t, right_t):
        op_str = f"{left} {op_sign} {right}"
        return MethodResult(op_str, bool)
    else:
        # replace = with == for comparisions
        raise _type_error_for_op(f"{op_sign_python_level or op_sign}", f"'{left_t}' and '{right_t}'")


def and_(prev_op, other):
    return _and_or(prev_op, other, "AND")


def or_(prev_op, other):
    return _and_or(prev_op, other, "OR")


def _and_or(prev_op, other, op_sign):
    left, right, left_t, right_t, op_str, _ = _init_binary_op(prev_op, other)
    if are_bools(left_t, right_t):
        op_str = f"{left} {op_sign} {right}"
        return MethodResult(op_str, prev_op.dtype)
    else:
        raise _type_error_for_op(f"{op_sign}", f"'{left_t}' and '{right_t}'")


def is_arithmetical(op):
    op = getattr(op, "_op_func", None)
    return op in {neg, pos, abs, add, sub, mul, div, mod}


def is_compare(op):
    op = getattr(op, "_op_func", None)
    return op in {invert, or_, and_, eq, ne, lt, le, gt, ge}



from functools import partial

import pandas as pd

import onetick.py as otp
from onetick.py.otq import otq

from onetick.py import utils
from onetick.py.sources.data_source import _start_doc, _end_doc, _symbol_doc
from onetick.py.docs.utils import docstring, param_doc


COMMON_SOURCE_DOC_PARAMS = [_start_doc, _end_doc, _symbol_doc]
OQD_TICK_TYPE = 'OQD::*'


def _modify_query_times(src):
    src.sink(otq.ModifyQueryTimes(
        start_time=('parse_time("%Y%m%d %H:%M:%S.%q", '
                    'time_format("%Y%m%d", _START_TIME, _TIMEZONE) + " 00:00:00.000", "GMT")'),
        end_time=('parse_time("%Y%m%d %H:%M:%S.%q", '
                  'time_format("%Y%m%d", _END_TIME, _TIMEZONE) + " 24:00:00.000", "GMT")'),
        output_timestamp='min(max(TIMESTAMP,_START_TIME),_END_TIME)'
    ))


_exch_doc = param_doc(
    name='exch',
    desc="""
    The OneQuantData exchange code for the desired price series. Possible values:

    - 'all'
        return data for all exchanges;

    - 'main'
        return data main pricing exchange;

    - any other string value will treated as exchange name to filter data.

    Default: 'all'.
    """,
    str_annotation="str, 'all', 'main'",
    default='all',
)


class OHLCV(otp.Source):
    @docstring(parameters=[_exch_doc] + COMMON_SOURCE_DOC_PARAMS, add_self=True)
    def __init__(
        self,
        exch='all',
        symbol=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
        **kwargs
    ):
        """
        OneQuantData™ source to retrieve a time series of unadjusted
        prices for a symbol for one particular pricing exchange of daily OHLCV data.
        Output ticks have fields: OPEN, HIGH, LOW, CLOSE, VOLUME, CURRENCY, EXCH.

        Examples
        --------
        >>> src = otp.oqd.sources.OHLCV(exch="USPRIM")  # doctest: +SKIP
        >>> otp.run(src,  # doctest: +SKIP
        ...         symbols='BTKR::::GOOGL US',
        ...         start=otp.dt(2018, 8, 1),
        ...         end=otp.dt(2018, 8, 2),
        ...         symbol_date=otp.dt(2018, 8, 1))
                         Time    OID    EXCH CURRENCY     OPEN     HIGH      LOW    CLOSE    VOLUME
        0 2018-08-01 00:00:00  74143  USPRIM      USD  1242.73  1245.72  1225.00  1232.99  605680.0
        1 2018-08-01 20:00:00  74143  USPRIM      USD  1219.69  1244.25  1218.06  1241.13  596960.0
        """
        if self._try_default_constructor(**kwargs):
            return

        super().__init__(
            _symbols=symbol,
            _start=start,
            _end=end,
            _base_ep_func=partial(self.build, exch=exch)
        )

        self.schema.set(OID=str,
                        EXCH=str,
                        CURRENCY=str,
                        OPEN=float,
                        HIGH=float,
                        LOW=float,
                        CLOSE=float,
                        VOLUME=float)

    def build(self, exch):
        ep = None
        if exch == 'all':
            ep = otp.oqd.eps.OqdSourceDprcAll()
        elif exch == 'main':
            ep = otp.oqd.eps.OqdSourceDprcMain()
        else:
            ep = otp.oqd.eps.OqdSourceDprcExch(exch=exch)

        src = otp.Source(ep)
        src.tick_type(OQD_TICK_TYPE)
        _modify_query_times(src)
        return src


class CorporateActions(otp.Source):
    """
    OneQuantData™ source EP to retrieve a time series of corporate
    actions for a symbol.

    This source will return all corporate action fields available for a symbol
    with EX-Dates between the query start time and end time.  The
    timestamp of the series is equal to the EX-Date of the corporate
    action with a time of 0:00:00 GMT.

    Examples
    --------
    >>> src = otp.oqd.sources.CorporateActions()  # doctest: +SKIP
    >>> otp.run(src,  # doctest: +SKIP
    ...         symbols='TDEQ::::AAPL',
    ...         start=otp.dt(2021, 1, 1),
    ...         end=otp.dt(2021, 8, 6),
    ...         symbol_date=otp.dt(2021, 2, 18),
    ...         timezone='GMT')
            Time   OID  ACTION_ID    ACTION_TYPE  ACTION_ADJUST ACTION_CURRENCY  ANN_DATE   EX_DATE  PAY_DATE  REC_DATE\
                       TERM_NOTE TERM_RECORD_TYPE ACTION_STATUS
    0 2021-02-05  9706   16799540  CASH_DIVIDEND          0.205             USD  20210127  20210205  20210211  20210208\
          CASH:0.205@USD                         NORMAL
    1 2021-05-07  9706   17098817  CASH_DIVIDEND          0.220             USD  20210428  20210507  20210513  20210510\
           CASH:0.22@USD                         NORMAL
    2 2021-08-06  9706   17331864  CASH_DIVIDEND          0.220             USD  20210727  20210806  20210812  20210809\
           CASH:0.22@USD                         NORMAL

    """

    @docstring(parameters=COMMON_SOURCE_DOC_PARAMS, add_self=True)
    def __init__(
        self,
        symbol=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
        **kwargs
    ):
        if self._try_default_constructor(**kwargs):
            return

        super().__init__(
            _symbols=symbol,
            _start=start,
            _end=end,
            _base_ep_func=partial(self.build)
        )

        self.schema.set(OID=str,
                        ACTION_ID=int,
                        ACTION_TYPE=str,
                        ACTION_ADJUST=float,
                        ACTION_CURRENCY=str,
                        ANN_DATE=int,
                        EX_DATE=int,
                        PAY_DATE=int,
                        REC_DATE=int,
                        TERM_NOTE=str,
                        TERM_RECORD_TYPE=str,
                        ACTION_STATUS=str)

    def build(self):
        ep = otp.oqd.eps.OqdSourceCacs()
        src = otp.Source(ep)
        src.tick_type(OQD_TICK_TYPE)
        _modify_query_times(src)
        return src


class DescriptiveFields(otp.Source):
    """OneQuantData™ source to retrieve a time series of descriptive fields for a symbol.
    There will only be ticks on days when some field in the descriptive data changes.
    Output ticks will have fields:
    OID, END_DATE, COUNTRY, EXCH, NAME,
    ISSUE_DESC, ISSUE_CLASS, ISSUE_TYPE, ISSUE_STATUS,
    SIC_CODE, IDSYM, TICKER, CALENDAR.

    Note: currently actual fields have 9999 year in END_DATE, but it could not fit the
    nanosecond timestamp, so it is replaced with 2035-01-01 date.

    Examples
    --------
    >>> src = otp.oqd.sources.DescriptiveFields()  # doctest: +SKIP
    >>> otp.run(src,  # doctest: +SKIP
    ...         symbols='1000001589',
    ...         start=otp.dt(2020, 3, 1),
    ...         end=otp.dt(2023, 3, 2),
    ...         timezone='GMT').iloc[:6]
            Time         OID    END_DATE COUNTRY  EXCH                NAME                   ISSUE_DESC\
                 ISSUE_CLASS ISSUE_TYPE ISSUE_STATUS SIC_CODE    IDSYM TICKER CALENDAR
    0 2020-03-01  1000001589  2020-03-23     LUX  EL^X  INVESTEC GLOBAL ST   EUROPEAN HIGH YLD BD INC 2\
                FUND                  NORMAL           B2PT4G9
    1 2020-03-23  1000001589  2020-04-01     LUX  EL^X  NINETY ONE LIMITED   EUROPEAN HIGH YLD BD INC 2\
                FUND                  NORMAL           B2PT4G9
    2 2020-04-01  1000001589  2021-01-01     LUX  EL^X  NINETY ONE LUX S.A   EUROPEAN HIGH YLD BD INC 2\
                FUND                  NORMAL           B2PT4G9
    3 2021-01-01  1000001589  2021-06-18     LUX  EL^X  NINETY ONE LUX S.A   EUROPEAN HIGH YLD BD INC 2\
                FUND                  NORMAL           B2PT4G9
    4 2021-06-18  1000001589  2022-01-01     LUX  EL^X  NINETY ONE LUX S.A  GSF GBL HIGH YLD A2 EUR DIS\
                FUND                  NORMAL           B2PT4G9
    5 2022-01-01  1000001589  2022-01-28     LUX  EL^X  NINETY ONE LUX S.A  GSF GBL HIGH YLD A2 EUR DIS\
                FUND                  NORMAL           B2PT4G9
    """

    @docstring(parameters=COMMON_SOURCE_DOC_PARAMS, add_self=True)
    def __init__(
        self,
        symbol=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
        **kwargs
    ):
        if self._try_default_constructor(**kwargs):
            return

        super().__init__(
            _symbols=symbol,
            _start=start,
            _end=end,
            _base_ep_func=partial(self.build)
        )

        self.schema.set(
            OID=str,
            END_DATE=otp.nsectime,
            COUNTRY=str,
            EXCH=str,
            NAME=str,
            ISSUE_DESC=str,
            ISSUE_CLASS=str,
            ISSUE_TYPE=str,
            ISSUE_STATUS=str,
            SIC_CODE=str,
            IDSYM=str,
            TICKER=str,
            CALENDAR=str,)

    def build(self):
        ep = otp.oqd.eps.OqdSourceDes()
        src = otp.Source(ep)
        src.tick_type(OQD_TICK_TYPE)
        _modify_query_times(src)

        # work-around to resolve problem with pandas timestamp out of bounds
        pd_max = pd.Timestamp.max.strftime('%Y%m%d%H%M%S')
        src.sink(otq.UpdateFields(
            set='END_DATE=PARSE_NSECTIME("%Y-%m-%d", "2035-01-01", _TIMEZONE)',
            where=f'AS_YYYYMMDDHHMMSS(END_DATE) > {pd_max}'))
        return src


class SharesOutstanding(otp.Source):
    """
    Logic is implemented in OQD_SOURCE_SHO EP to retrieve a time series of shares
    outstanding for a stock.

    The source retrieves a time series of shares outstanding
    for a stock. This source only applies to stocks or securities that have
    published shares outstanding data.

    The series represents total shares outstanding and is not free float
    adjusted.

    Note: currently actual fields have 9999 year in END_DATE, but it could not fit the
    nanosecond timestamp, so it is replaced with 2035-01-01 date.


    Examples
    --------
    >>> src = otp.oqd.sources.SharesOutstanding()  # doctest: +SKIP
    >>> otp.run(src,  # doctest: +SKIP
    ...         symbols='TDEQ::::AAPL',
    ...         start=otp.dt(2021, 1, 1),
    ...         end=otp.dt(2021, 8, 6),
    ...         symbol_date=otp.dt(2021, 2, 18),
    ...         timezone='GMT')
            Time   OID   END_DATE REPORT_MONTH        SHARES
    0 2021-01-01  9706 2021-01-06       202009  1.700180e+10
    1 2021-01-06  9706 2021-01-29       202009  1.682326e+10
    2 2021-01-29  9706 2021-05-03       202012  1.678810e+10
    3 2021-05-03  9706 2021-07-30       202103  1.668763e+10
    4 2021-07-30  9706 2021-10-29       202106  1.653017e+10
    """
    @docstring(parameters=COMMON_SOURCE_DOC_PARAMS, add_self=True)
    def __init__(
        self,
        symbol=otp.utils.adaptive,
        start=otp.utils.adaptive,
        end=otp.utils.adaptive,
        **kwargs
    ):
        if self._try_default_constructor(**kwargs):
            return

        super().__init__(
            _symbols=symbol,
            _start=start,
            _end=end,
            _base_ep_func=partial(self.build)
        )

        self.schema.set(OID=str,
                        END_DATE=otp.nsectime,
                        REPORT_MONTH=int,
                        SHARES=int)

    def build(self):
        ep = otp.oqd.eps.OqdSourceSho()
        src = otp.Source(ep)
        src.tick_type(OQD_TICK_TYPE)
        _modify_query_times(src)

        # work-around to resolve problem with pandas timestamp out of bounds
        pd_max = pd.Timestamp.max.strftime('%Y%m%d%H%M%S')
        src.sink(otq.UpdateFields(
            set='END_DATE=PARSE_NSECTIME("%Y-%m-%d", "2035-01-01", _TIMEZONE)',
            where=f'AS_YYYYMMDDHHMMSS(END_DATE) > {pd_max}'))
        return src



import json

from onetick.py.otq import otq, pyomd
from datetime import datetime


def get_symbol_list_from_df(df, symbol_name_column='SYMBOL_NAME'):
    """
    Creates a onetick.query.Symbol object that may be passed as a symbol list from the dataframe
    with query results. SYMBOL_NAME column is interpreted as symbol names, while other columns are
    interpreted as symbol params.
    """
    if symbol_name_column not in df.columns:
        raise ValueError(f'Dataframe used as symbol list does not contain a {symbol_name_column} column')

    df = df.drop(columns=['Time'])

    def symbol_from_dict(params):
        name = params[symbol_name_column]
        del params[symbol_name_column]
        return otq.Symbol(name=name, params=params)

    symbols = [symbol_from_dict(row) for row in df.to_dict(orient='records')]
    return symbols


class JSONEncoder(json.JSONEncoder):
    """
    onetick.py json encoder that also supports some of the onetick.py objects like otp.adaptive.
    """
    def default(self, o):
        # let's use python str representation by default, all objects in python should have it
        # maybe we will add better representations for some types later
        return str(o)


def json_dumps(obj, **kwargs) -> str:
    """
    Wrapper around json.dumps that also supports some of the onetick.py objects like otp.adaptive.
    ``kwargs`` arguments are propagated to json.dumps function.
    By default ``cls`` parameter is set to otp.utils.JSONEncoder.
    """
    kwargs.setdefault('cls', JSONEncoder)
    return json.dumps(obj, **kwargs)


def query_properties_to_dict(query_properties: pyomd.QueryProperties) -> dict:  # type: ignore[valid-type]
    """
    Convert :py:class:`pyomd.QueryProperties` to dictionary.
    """

    str_qp = query_properties.convert_to_name_value_pairs_string()  # type: ignore[attr-defined]
    if not isinstance(str_qp, str):
        str_qp = str_qp.c_str()
    pairs = str_qp.split(',') if str_qp else []
    return dict(pair.split('=', maxsplit=1) for pair in pairs)


def query_properties_from_dict(query_properties_dict: dict) -> pyomd.QueryProperties:  # type: ignore[valid-type]
    """
    Convert dictionary to :py:class:`pyomd.QueryProperties`.
    """
    query_properties = pyomd.QueryProperties()
    for k, v in query_properties_dict.items():
        query_properties.set_property_value(k, v)
    return query_properties


def symbol_date_to_str(symbol_date) -> str:
    if isinstance(symbol_date, int):
        symbol_date = str(symbol_date)
    if isinstance(symbol_date, str):
        symbol_date = datetime.strptime(symbol_date, '%Y%m%d')
    if hasattr(symbol_date, 'strftime'):
        symbol_date = symbol_date.strftime('%Y%m%d')
    return symbol_date



from typing import Dict
import collections.abc


class Schema(collections.abc.Mapping):
    """
    A source data schema proxy. It allows to work with source schema in
    a frozen-dict manner. It also allows to set using the ``set``
    and update a source schema using the ``update`` methods.
    """

    dict_cls = dict

    def __init__(self, _base_source=None, _hidden_columns=None, **kwargs):
        self._dict = self.dict_cls(**kwargs)
        self._base_source = _base_source
        self._hash = None
        self._hidden_columns = _hidden_columns if _hidden_columns else {}

    def __getitem__(self, item):
        if isinstance(item, list):
            return {key: value for key, value in self._dict.items()
                    if key in item}

        if item in self._hidden_columns:
            return self._hidden_columns[item]

        return self._dict[item]

    def __setitem__(self, key, value):
        self.update(**{key: value})

    def __contains__(self, item):
        if item in self._hidden_columns:
            return True

        return item in self._dict

    def __iter__(self):
        return iter(self._dict)

    def items(self):
        return self._dict.items()

    def keys(self):
        return self._dict.keys()

    def copy(self):
        return self._dict.copy()

    def set(self, **new_schema: type):
        """
        Drops the python schema representation of a source and sets the new one from the `new_schema`

        Parameters
        ----------
        new_schema: Dict[str, type]
            schema in the column-name -> type format

        Returns
        -------
            None
        """
        self._base_source.set_schema(**new_schema)

    def update(self, **other_schema: type):
        """
        Updates the python schema representation of a source: values from matching keys
        will be overridden from the `other_schema`, values from new keys will be added.

        Parameters
        ----------
        other_schema:
            schema in the column-name -> type format

        Returns
        -------
            None
        """
        current_schema = self._base_source.columns(skip_meta_fields=True)
        current_schema.update(other_schema)

        self._base_source.set_schema(**current_schema)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        return repr(self._dict)

    def __hash__(self):
        if self._hash is None:
            h = 0
            for key, value in self._dict.items():
                h ^= hash((key, value))
            self._hash = h
        return self._hash



import pandas as pd
import numpy as np

from .. import types as ott


def _convert_pandas_types(dtype):
    if pd.api.types.is_integer_dtype(dtype):
        return int
    elif pd.api.types.is_float_dtype(dtype):
        return float
    elif pd.api.types.is_string_dtype(dtype):
        return str
    elif pd.api.types.is_bool_dtype(dtype):
        return bool
    return None


def inspect_by_pandas(path_to_csv, first_line_is_title=True, names=None, field_delimiter=',', quote_char='"'):
    # read CHUNK_SIZE first lines to determine column types
    CHUNK_SIZE = 300

    header = None

    if first_line_is_title:
        header = 0 if names else "infer"

    with pd.read_csv(
        path_to_csv,
        engine="python",
        iterator=True,
        header=header,
        names=names,
        sep=field_delimiter,
        escapechar='\\',
        quotechar=quote_char,
    ) as reader:
        df = reader.read(CHUNK_SIZE)
        prefix = None if first_line_is_title or names else "COLUMN_"
        if prefix:
            df.columns = [f'{prefix}{col}' for col in df.columns]

    if not first_line_is_title:
        first_column = "COLUMN_0"
        if names:
            first_column = names[0]
        if len(df) > 0 and len(df.columns) > 0 and df.dtypes[first_column] == np.dtype("O"):
            if df[first_column][0].startswith("#"):
                raise ValueError(
                    "If first line of CSV starts with #, you must set first_line_is_title=True, "
                    "because OneTick will forcefully use first line as header.")

    # CSV_FILE_LISTING will ignore FIRST_LINE_IS_TITLE, if first line starts with hash sign #
    forced_title = df.columns.values[0][0] == '#'
    if forced_title:
        # remove hash sign from first column name
        df.rename(columns={df.columns.values[0]: df.columns.values[0][1:]}, inplace=True)

    # check for default types in OneTick format ("columnname type")
    default_types = {}
    rename = {}

    for column in df.columns.values:
        c = column.split()

        if column.startswith('Unnamed: '):
            # OneTick doesn't allow to have empty column name, so we explicitly set it here
            # and in CSV_FILE_LISTING EP later in CSV.base_ep()
            rename[column] = "COLUMN_" + column[9:]

        elif len(c) == 2:
            # format: "type COLUMNNAME"
            # http://solutions.pages.soltest.onetick.com/iac/onetick-server/ep_guide/EP/FieldTypeDeclarations.htm#supported_field_types
            dtype = ott.str2type(c[0])
            if dtype is not None:
                default_types[c[1]] = dtype
                rename[column] = c[1]

    if rename:
        df.rename(columns=rename, inplace=True)

    # convert pandas types to otp
    columns = dict(map(lambda x: (x[0], _convert_pandas_types(x[1])), dict(df.dtypes).items()))

    # explicitly set types for columns having format "type COLUMNNAME"
    for column_name, dtype in default_types.items():
        columns[column_name] = dtype

    # reset default_types if # is not in first line
    if not forced_title:
        default_types = {}

    return columns, default_types, forced_title



import onetick.py as otp
import onetick.py.types as ott
from onetick.py.core.column import _Column
from onetick.py.core.column_operations.base import _Operation


class _SymbolParamColumn(_Column):
    """
    Internal object representing OneTick's symbol parameters.
    Can be used in other onetick.py methods to specify if object is symbol parameter.
    """

    def __init__(self, *args, default=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._default = default

    def __str__(self):
        result = super().__str__()

        # TODO: PY-35
        # This is ad-hoc check, really we need to change column formatting to
        # pre- and post-formats, and copy columns through the .copy() method
        # on the _Column instead of copying them manually in different places
        # of the _Source class
        if self.name != '_SYMBOL_NAME':
            result = f'_SYMBOL_PARAM.{self.name}'

        if self._default is not None:
            default_formatted = ott.value2str(self._default)
            default_formatted = f'TOSTRING({default_formatted})'
            result = f'CASE(UNDEFINED("{result}"), false, {result}, {default_formatted})'

        # symbol params are always string, need to convert
        result = _Operation(op_str=result, dtype=str)

        if self.dtype is ott.nsectime:
            result = result.str.to_datetime(unit='ns')
        elif self.dtype is ott.msectime:
            result = result.str.to_datetime(unit='ms')
        else:
            result = result.astype(self.dtype)

        return str(result)


class _SymbolParamSource:
    """
    Internal container that provides access to symbol parameters.
    The object is read-only, you can only get symbol parameters with it.
    """

    def __init__(self, **columns):

        if 'name' in columns:
            raise ValueError(
                'name field is a specific field for accessing _SYMBOL_NAME variable, please rename your parameter'
            )

        # "SYMBOL_NAME" field won't be propagated as a symbol parameter by onetick, so we should remove it here
        columns.pop('SYMBOL_NAME', None)

        for key, dtype in columns.items():
            # TODO: change to immutable Column, PY-35
            self.__dict__[key] = _SymbolParamColumn(key, dtype, self)

        self.__dict__['name'] = _SymbolParamColumn('_SYMBOL_NAME', str, self)
        self.__dict__['_SYMBOL_NAME'] = self.__dict__['name']

        # set schema
        schema = columns.copy()

        for meta_c in otp.meta_fields.get_onetick_fields_and_types():
            schema.pop(meta_c, None)

        self.__dict__['_schema'] = schema

    @property
    def schema(self):
        """
        Represents actual python data schema in the column name -> type format.
        """
        return self._schema

    def __getitem__(self, key):
        """
        Get symbol parameter with corresponding ``key``.
        Raises an error if such column does not exist.
        """
        if key not in self.__dict__:
            raise AttributeError(f"There is no '{key}' column")

        return self.__dict__[key]

    def __setattr__(self, key, value):
        raise NotImplementedError('Symbol params are read only')



from typing import TYPE_CHECKING, List, Optional, Union
from onetick.py.backports import Literal

from abc import ABC

import onetick.py as otp
from onetick.py.otq import otq
from onetick.py import types as ott

if TYPE_CHECKING:
    from onetick.py.core.source import Source   # hack for annotations
from onetick.py.core.column import _Column
from onetick.py.compatibility import is_supported_otq_ob_summary
from ._base import _Aggregation, get_seconds_from_time_offset
from ._docs import (_running_doc,
                    _bucket_interval_doc,
                    _bucket_time_doc,
                    _bucket_units_ob_doc,
                    _bucket_end_condition_doc,
                    _end_condition_per_group_doc,
                    _group_by_doc,
                    _groups_to_display_doc,
                    _side_doc,
                    _max_levels_doc,
                    _min_levels_doc,
                    _max_depth_shares_doc,
                    _max_depth_for_price_doc,
                    _book_uncross_method_doc,
                    _dq_events_that_clear_book_doc,
                    _best_ask_price_field_doc,
                    _best_bid_price_field_doc,
                    _bucket_interval_ob_num_levels_doc,
                    _identify_source_doc,
                    _show_full_detail_doc,
                    _show_only_changes_doc,
                    _book_delimiters_doc,
                    _max_initialization_days_doc,
                    _state_key_max_inactivity_sec_doc,
                    _size_max_fractional_digits_doc,
                    _include_market_order_ticks_doc)


OB_SNAPSHOT_DOC_PARAMS = [
    _running_doc,
    _bucket_interval_doc, _bucket_time_doc, _bucket_units_ob_doc,
    _bucket_end_condition_doc, _end_condition_per_group_doc, _group_by_doc, _groups_to_display_doc,
    _side_doc, _max_levels_doc, _max_depth_shares_doc, _max_depth_for_price_doc,
    _book_uncross_method_doc, _dq_events_that_clear_book_doc, _identify_source_doc,
    _show_full_detail_doc, _show_only_changes_doc, _book_delimiters_doc,
    _max_initialization_days_doc, _state_key_max_inactivity_sec_doc,
    _size_max_fractional_digits_doc,
    _include_market_order_ticks_doc,
]
OB_SNAPSHOT_WIDE_DOC_PARAMS = [
    _running_doc,
    _bucket_interval_doc, _bucket_time_doc, _bucket_units_ob_doc, _bucket_end_condition_doc,
    _end_condition_per_group_doc, _group_by_doc, _groups_to_display_doc,
    _max_levels_doc, _max_depth_shares_doc, _max_depth_for_price_doc,
    _book_uncross_method_doc, _dq_events_that_clear_book_doc,
    _book_delimiters_doc,
    _max_initialization_days_doc, _state_key_max_inactivity_sec_doc,
    _size_max_fractional_digits_doc,
    _include_market_order_ticks_doc,
]
OB_SNAPSHOT_FLAT_DOC_PARAMS = [
    _running_doc,
    _bucket_interval_doc, _bucket_time_doc, _bucket_units_ob_doc, _bucket_end_condition_doc,
    _end_condition_per_group_doc, _group_by_doc, _groups_to_display_doc,
    _max_levels_doc,
    _book_uncross_method_doc, _dq_events_that_clear_book_doc,
    _show_full_detail_doc,
    _max_initialization_days_doc, _state_key_max_inactivity_sec_doc,
    _size_max_fractional_digits_doc,
    _include_market_order_ticks_doc,
]
OB_SUMMARY_DOC_PARAMS = [
    _running_doc,
    _bucket_interval_doc, _bucket_time_doc, _bucket_units_ob_doc,
    _bucket_end_condition_doc, _end_condition_per_group_doc, _group_by_doc, _groups_to_display_doc,
    _side_doc, _max_levels_doc, _min_levels_doc, _max_depth_shares_doc, _max_depth_for_price_doc,
    _book_uncross_method_doc, _dq_events_that_clear_book_doc, _max_initialization_days_doc,
    _state_key_max_inactivity_sec_doc, _size_max_fractional_digits_doc,
    _include_market_order_ticks_doc,
]

OB_SIZE_DOC_PARAMS = [
    _running_doc,
    _bucket_interval_doc, _bucket_time_doc, _bucket_units_ob_doc,
    _bucket_end_condition_doc, _end_condition_per_group_doc, _group_by_doc, _groups_to_display_doc,
    _side_doc, _max_levels_doc, _max_depth_for_price_doc,
    _book_uncross_method_doc, _dq_events_that_clear_book_doc, _max_initialization_days_doc,
    _best_ask_price_field_doc, _best_bid_price_field_doc,
]

OB_VWAP_DOC_PARAMS = [
    _running_doc,
    _bucket_interval_doc, _bucket_time_doc, _bucket_units_ob_doc,
    _bucket_end_condition_doc, _end_condition_per_group_doc, _group_by_doc, _groups_to_display_doc,
    _side_doc, _max_levels_doc, _max_depth_shares_doc,
    _book_uncross_method_doc, _dq_events_that_clear_book_doc, _max_initialization_days_doc,
]

OB_NUM_LEVELS_DOC_PARAMS = [
    _running_doc,
    _bucket_interval_ob_num_levels_doc, _side_doc,
    _book_uncross_method_doc, _dq_events_that_clear_book_doc, _max_initialization_days_doc,
]


class _OrderBookAggregation(_Aggregation, ABC):
    FIELDS_TO_SKIP = ['column_name', 'all_fields', 'boundary_tick_bucket', 'output_field_name']
    FIELDS_MAPPING = dict(_Aggregation.FIELDS_MAPPING, **{
        'side': 'SIDE',
        'max_levels': 'MAX_LEVELS',
        'max_depth_shares': 'MAX_DEPTH_SHARES',
        'max_depth_for_price': 'MAX_DEPTH_FOR_PRICE',
        'max_initialization_days': 'MAX_INITIALIZATION_DAYS',
        'book_uncross_method': 'BOOK_UNCROSS_METHOD',
        'dq_events_that_clear_book': 'DQ_EVENTS_THAT_CLEAR_BOOK',
    })
    FIELDS_DEFAULT = dict(_Aggregation.FIELDS_DEFAULT, **{
        'side': None,
        'max_levels': None,
        'max_depth_shares': None,
        'max_depth_for_price': None,
        'max_initialization_days': 1,
        'book_uncross_method': None,
        'dq_events_that_clear_book': None,
    })
    _validations_to_skip = ['running_all_fields']

    def __init__(self,
                 *args,
                 side: Optional[Literal['ASK', 'BID']] = None,
                 max_levels: Optional[int] = None,
                 max_depth_shares: Optional[int] = None,
                 max_depth_for_price: Optional[float] = None,
                 max_initialization_days: int = 1,
                 book_uncross_method: Optional[Literal['REMOVE_OLDER_CROSSED_LEVELS']] = None,
                 dq_events_that_clear_book: Optional[List[str]] = None,
                 **kwargs):
        self.side = side
        self.max_levels = max_levels
        self.max_depth_shares = max_depth_shares
        self.max_depth_for_price = max_depth_for_price
        self.max_initialization_days = max_initialization_days
        self.book_uncross_method = book_uncross_method
        self.dq_events_that_clear_book = ','.join(dq_events_that_clear_book) if dq_events_that_clear_book else None
        super().__init__(_Column('TIMESTAMP'), *args, **kwargs)

    def _param_validation(self):
        super()._param_validation()
        book_uncross_methods = (None, 'REMOVE_OLDER_CROSSED_LEVELS')
        if self.book_uncross_method not in book_uncross_methods:
            raise ValueError(
                f"Wrong value for parameter 'book_uncross_method': '{self.book_uncross_method}'. "
                f"Possible values: {book_uncross_methods}."
            )
        valid_units = ("seconds", "days", "months", "flexible")
        if self.bucket_units not in valid_units:
            raise ValueError("'bucket_units' can be one of the following: "
                             f"'{', '.join(valid_units)}'; however, '{self.bucket_units}' was passed")

    def validate_input_columns(self, src: 'Source'):
        super().validate_input_columns(src)
        if any([
            not {'BUY_SELL_FLAG', 'PRICE', 'SIZE'}.issubset(src.schema),
            'UPDATE_TIME' not in src.schema and 'DELETED_TIME' not in src.schema
        ]):
            raise TypeError(f"Aggregation `{self.NAME}` need these columns: "
                            f"BUY_SELL_FLAG, PRICE, SIZE and (UPDATE_TIME or DELETED_TIME)")


class ObSnapshot(_OrderBookAggregation):
    NAME = 'OB_SNAPSHOT'
    EP = otq.ObSnapshot

    FIELDS_MAPPING = dict(_OrderBookAggregation.FIELDS_MAPPING, **{
        'identify_source': 'IDENTIFY_SOURCE',
        'show_full_detail': 'SHOW_FULL_DETAIL',
        'show_only_changes': 'SHOW_ONLY_CHANGES',
        'book_delimiters': 'BOOK_DELIMITERS',
        'state_key_max_inactivity_sec': 'STATE_KEY_MAX_INACTIVITY_SEC',
        'size_max_fractional_digits': 'SIZE_MAX_FRACTIONAL_DIGITS',
        'include_market_order_ticks': 'INCLUDE_MARKET_ORDER_TICKS',
    })
    FIELDS_DEFAULT = dict(_OrderBookAggregation.FIELDS_DEFAULT, **{
        'identify_source': False,
        'show_full_detail': False,
        'show_only_changes': False,
        'book_delimiters': None,
        'state_key_max_inactivity_sec': None,
        'size_max_fractional_digits': 0,
        'include_market_order_ticks': None,
    })

    def __init__(self,
                 *args,
                 identify_source: bool = False,
                 show_full_detail: bool = False,
                 show_only_changes: bool = False,
                 book_delimiters: Optional[Literal['D']] = None,
                 state_key_max_inactivity_sec: Optional[int] = None,
                 size_max_fractional_digits: int = 0,
                 include_market_order_ticks: Optional[bool] = None,
                 **kwargs):
        self.identify_source = identify_source
        self.show_full_detail = show_full_detail
        self.show_only_changes = show_only_changes
        self.book_delimiters = book_delimiters
        self.state_key_max_inactivity_sec = state_key_max_inactivity_sec
        self.size_max_fractional_digits = size_max_fractional_digits
        self.include_market_order_ticks = include_market_order_ticks
        # we don't want to set hard limit on the output of order book aggregations
        if self.show_full_detail:
            kwargs['all_fields'] = True
        super().__init__(*args, **kwargs)

    def _param_validation(self):
        super()._param_validation()
        if self.include_market_order_ticks is not None:
            if 'include_market_order_ticks' not in self.EP.Parameters.list_parameters():
                raise ValueError("Parameter 'include_market_order_ticks' is not supported on this OneTick API version")
            otp.compatibility.is_include_market_order_ticks_supported(
                throw_warning=True,
                feature_name="parameter 'include_market_order_ticks'",
            )

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None) -> dict:
        schema = {
            'PRICE': float,
            'SIZE': int,
            'LEVEL': int,
            'UPDATE_TIME': otp.nsectime,
            'BUY_SELL_FLAG': int,
        }
        if self.book_delimiters:
            schema['DELIMITER'] = str
        return schema


class ObSnapshotWide(ObSnapshot):
    NAME = 'OB_SNAPSHOT_WIDE'
    EP = otq.ObSnapshotWide

    FIELDS_TO_SKIP = [
        *ObSnapshot.FIELDS_TO_SKIP, 'side', 'identify_source', 'show_full_detail', 'show_only_changes'
    ]

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None) -> dict:
        schema = {
            'BID_PRICE': float,
            'BID_SIZE': int,
            'BID_UPDATE_TIME': otp.nsectime,
            'ASK_PRICE': float,
            'ASK_SIZE': int,
            'ASK_UPDATE_TIME': otp.nsectime,
            'LEVEL': int,
        }
        if self.book_delimiters:
            schema['DELIMITER'] = str
        return schema


class ObSnapshotFlat(ObSnapshot):
    NAME = 'OB_SNAPSHOT_FLAT'
    EP = otq.ObSnapshotFlat

    FIELDS_TO_SKIP = [
        *ObSnapshot.FIELDS_TO_SKIP,
        'side', 'identify_source', 'show_only_changes',
        'book_delimiters', 'max_depth_shares', 'max_depth_for_price',
    ]

    def validate_input_columns(self, src: 'Source'):
        super().validate_input_columns(src)
        if self.max_levels is None or self.max_levels < 1 or self.max_levels > 100_000:
            raise ValueError(f"Parameter 'max_levels' must be set in aggregation `{self.NAME}`"
                             f" and must be between 1 and 100000.")

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None) -> dict:
        schema = {}
        assert self.max_levels is not None
        for level in range(1, self.max_levels + 1):
            schema.update({
                f'BID_PRICE{level}': float,
                f'BID_SIZE{level}': int,
                f'BID_UPDATE_TIME{level}': otp.nsectime,
                f'ASK_PRICE{level}': float,
                f'ASK_SIZE{level}': int,
                f'ASK_UPDATE_TIME{level}': otp.nsectime,
            })
        return schema


class ObSummary(_OrderBookAggregation):
    NAME = 'OB_SUMMARY'

    # Will be set later, to prevent error while importing this module with outdated onetick
    EP = None

    FIELDS_MAPPING = dict(_OrderBookAggregation.FIELDS_MAPPING, **{
        'min_levels': 'MIN_LEVELS',
        'state_key_max_inactivity_sec': 'STATE_KEY_MAX_INACTIVITY_SEC',
        'size_max_fractional_digits': 'SIZE_MAX_FRACTIONAL_DIGITS',
        'include_market_order_ticks': 'INCLUDE_MARKET_ORDER_TICKS',
    })
    FIELDS_DEFAULT = dict(_OrderBookAggregation.FIELDS_DEFAULT, **{
        'min_levels': None,
        'state_key_max_inactivity_sec': None,
        'size_max_fractional_digits': 0,
        'include_market_order_ticks': None,
    })

    def __init__(self,
                 *args,
                 min_levels: Optional[int] = None,
                 state_key_max_inactivity_sec: Optional[int] = None,
                 size_max_fractional_digits: int = 0,
                 include_market_order_ticks: Optional[bool] = None,
                 **kwargs):
        if is_supported_otq_ob_summary():
            self.EP = otq.ObSummary
        else:
            raise RuntimeError("Used onetick installation not support onetick.query.ObSummary")

        self.min_levels = min_levels
        self.state_key_max_inactivity_sec = state_key_max_inactivity_sec
        self.size_max_fractional_digits = size_max_fractional_digits
        self.include_market_order_ticks = include_market_order_ticks
        super().__init__(*args, **kwargs)

    def _param_validation(self):
        super()._param_validation()
        if self.include_market_order_ticks is not None:
            if 'include_market_order_ticks' not in self.EP.Parameters.list_parameters():
                raise ValueError("Parameter 'include_market_order_ticks' is not supported on this OneTick API version")
            otp.compatibility.is_include_market_order_ticks_supported(
                throw_warning=True,
                feature_name="parameter 'include_market_order_ticks'",
            )

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None) -> dict:
        schema = {
            'BID_SIZE': int,
            'BID_VWAP': float,
            'BEST_BID_PRICE': float,
            'WORST_BID_PRICE': float,
            'NUM_BID_LEVELS': int,
            'ASK_SIZE': int,
            'ASK_VWAP': float,
            'BEST_ASK_PRICE': float,
            'WORST_ASK_PRICE': float,
            'NUM_ASK_LEVELS': int,
        }
        return schema


class ObSize(_OrderBookAggregation):
    NAME = 'OB_SIZE'
    EP = otq.ObSize

    FIELDS_MAPPING = dict(_OrderBookAggregation.FIELDS_MAPPING, **{
        'min_levels': 'MIN_LEVELS',
        'best_ask_price_field': 'BEST_ASK_PRICE_FIELD',
        'best_bid_price_field': 'BEST_BID_PRICE_FIELD',
    })
    FIELDS_DEFAULT = dict(_OrderBookAggregation.FIELDS_DEFAULT, **{
        'min_levels': None,
        'best_ask_price_field': '',
        'best_bid_price_field': '',
    })

    def __init__(
        self, *args, min_levels: Optional[int] = None,
        best_ask_price_field: Optional[Union[str, _Column]] = None,
        best_bid_price_field: Optional[Union[str, _Column]] = None,
        **kwargs,
    ):
        if min_levels and not kwargs.get('max_depth_for_price'):
            raise ValueError('`min_levels` parameter must not be set when `max_depth_for_price` not set')

        self.min_levels = min_levels

        if isinstance(best_ask_price_field, _Column):
            best_ask_price_field = str(best_ask_price_field)
        elif best_ask_price_field is None:
            best_ask_price_field = ''

        if isinstance(best_bid_price_field, _Column):
            best_bid_price_field = str(best_bid_price_field)
        elif best_bid_price_field is None:
            best_bid_price_field = ''

        self.best_ask_price_field = best_ask_price_field
        self.best_bid_price_field = best_bid_price_field
        super().__init__(*args, **kwargs)

    def validate_input_columns(self, src: 'Source'):
        super().validate_input_columns(src)

        if self.best_ask_price_field and self.best_ask_price_field not in src.schema:
            raise ValueError(
                f'Column \'{self.best_ask_price_field}\' from `best_ask_price_field` parameter not in the schema.'
            )

        if self.best_bid_price_field and self.best_bid_price_field not in src.schema:
            raise ValueError(
                f'Column \'{self.best_bid_price_field}\' from `best_bid_price_field` parameter not in the schema.'
            )

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None) -> dict:
        if self.side:
            return {'VALUE': float}

        return {
            'ASK_VALUE': float,
            'BID_VALUE': float,
        }


class ObVwap(_OrderBookAggregation):
    NAME = 'OB_VWAP'
    EP = otq.ObVwap

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None) -> dict:
        if self.side:
            return {'VALUE': float}

        return {
            'ASK_VALUE': float,
            'BID_VALUE': float,
        }


class ObNumLevels(_OrderBookAggregation):
    NAME = 'OB_NUM_LEVELS'
    EP = otq.ObNumLevels

    def __init__(self, *args, bucket_interval: Union[int, ott.OTPBaseTimeOffset] = 0, **kwargs):
        if not isinstance(bucket_interval, (int, ott.OTPBaseTimeOffset)):
            raise ValueError('Unsupported value type for `bucket_interval` parameter')

        if isinstance(bucket_interval, ott.OTPBaseTimeOffset):
            _, datepart = bucket_interval.get_offset()
            if datepart not in {'second', 'minute', 'hour', 'day'}:
                raise ValueError(f"Unsupported DatePart passed to bucket_interval: {datepart}")

            bucket_interval = get_seconds_from_time_offset(bucket_interval)

        super().__init__(*args, bucket_interval=bucket_interval, **kwargs)

    def _get_output_schema(self, src: 'Source', name: Optional[str] = None) -> dict:
        if self.side:
            return {'VALUE': float}

        return {
            'ASK_VALUE': float,
            'BID_VALUE': float,
        }



import onetick.py as otp
from onetick.py.otq import otq

from onetick.py.core.source import Source

from .. import utils

from .common import update_node_tick_type


class ReadSnapshot(Source):
    def __init__(
        self,
        snapshot_name='VALUE',
        snapshot_storage='memory',
        allow_snapshot_absence=False,
        symbol=utils.adaptive,
        db=utils.adaptive_to_default,
        tick_type=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
        schema=None,
        **kwargs,
    ):
        """
        Reads ticks for a specified symbol name and snapshot name from global memory storage or
        from a memory mapped file.

        These ticks should be written there by the :py:meth:`onetick.py.Source.save_snapshot` event processor.
        Ticks with an empty symbol name (for example, those after merging the time series of different symbols)
        are saved under **CEP_SNAPSHOT::** symbol name, so for reading such ticks a dummy database
        with the name **CEP_SNAPSHOT** must be configured in the database locator file.

        Parameters
        ----------
        snapshot_name: str
            The name that was specified in :py:meth:`onetick.py.Source.save_snapshot` as a ``snapshot_name``
            during saving.

            Default: ``VALUE``
        snapshot_storage: 'memory' or 'memory_mapped_file'
            This parameter specifies the place of storage of the snapshot. Possible options are:

            * `memory` - the snapshot is stored in the dynamic (heap) memory of the process
              that ran (or is still running) the :py:meth:`onetick.py.Source.save_snapshot` for the snapshot.
            * `memory_mapped_file` - the snapshot is stored in a memory mapped file.
              For each symbol to get the location of the snapshot in the file system, ``ReadSnapshot`` looks at
              the **SAVE_SNAPSHOT_DIR** parameter value in the locator section for the database of the symbol.

            Default: `memory`
        allow_snapshot_absence: bool
            If specified, the EP does not display an error about missing snapshot
            if the snapshot has not been saved or is still being saved.

            Default: `False`
        symbol: str, list of str, :class:`Source`, :class:`query`, :py:func:`eval query <onetick.py.eval>`
            Symbol(s) from which data should be taken.
        tick_type: str
            Tick type.
            Default: ANY.
        start: :py:class:`otp.datetime <onetick.py.datetime>`
            Start time for tick generation. By default the start time of the query will be used.
        end: :py:class:`otp.datetime <onetick.py.datetime>`
            End time for tick generation. By default the end time of the query will be used.
        schema: dict
            Dictionary of columns names with their types.

            .. warning::
                You should set schema manually, if you want to use fields in `onetick-py` query description
                before its execution.

        See also
        --------
        | **READ_SNAPSHOT** OneTick event processor
        | :py:class:`onetick.py.ShowSnapshotList`
        | :py:class:`onetick.py.FindSnapshotSymbols`
        | :py:meth:`onetick.py.Source.save_snapshot`
        | :py:meth:`onetick.py.Source.join_with_snapshot`

        Examples
        --------
        Read snapshot from memory:

        >>> src = otp.ReadSnapshot(snapshot_name='some_snapshot')
        >>> otp.run(src)  # doctest: +SKIP
                Time  PRICE  SIZE               TICK_TIME
        0 2003-12-01  100.2   500 2003-12-01 00:00:00.000
        1 2003-12-01   98.3   250 2003-12-01 00:00:00.001
        2 2003-12-01  102.5   400 2003-12-01 00:00:00.002

        You can specify schema manually in order to reference snapshot fields while constructing query via `onetick-py`:

        >>> src = otp.ReadSnapshot(
        ...     snapshot_name='some_snapshot', schema={'PRICE': float, 'SIZE': int},
        ... )
        >>> src['VOLUME'] = src['PRICE'] * src['SIZE']

        Read snapshot for specified database and symbol name:

        >>> src = otp.ReadSnapshot(snapshot_name='some_snapshot', db='DB', symbol='AAA')

        Read snapshot from memory mapped file:

        >>> src = otp.ReadSnapshot(
        ...     snapshot_name='some_snapshot', snapshot_storage='memory_mapped_file', db='DB',
        ... )
        """
        if self._try_default_constructor(schema=schema):
            return

        if not hasattr(otq, "ReadSnapshot"):
            raise RuntimeError("Current version of OneTick don't support READ_SNAPSHOT EP")

        if schema is None:
            schema = {}

        schema.update({'TICK_TIME': otp.nsectime})

        if snapshot_storage not in ['memory', 'memory_mapped_file']:
            raise ValueError('`snapshot_storage` must be one of "memory", "memory_mapped_file"')

        super().__init__(
            _symbols=symbol,
            _start=start,
            _end=end,
            _base_ep_func=lambda: self.base_ep(
                db=db,
                tick_type=tick_type,
                snapshot_name=snapshot_name,
                snapshot_storage=snapshot_storage,
                allow_snapshot_absence=allow_snapshot_absence,
            ),
            schema=schema,
            **kwargs,
        )

    def base_ep(
        self,
        snapshot_name='VALUE',
        snapshot_storage='memory',
        allow_snapshot_absence=False,
        db=utils.adaptive_to_default,
        tick_type=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
    ):
        snapshot_storage = snapshot_storage.upper()

        src = Source(
            otq.ReadSnapshot(
                snapshot_name=snapshot_name,
                snapshot_storage=snapshot_storage,
                allow_snapshot_absence=allow_snapshot_absence,
            )
        )

        if db or tick_type:
            update_node_tick_type(src, tick_type, db)

        return src


class ShowSnapshotList(Source):
    def __init__(self, snapshot_storage='all', **kwargs):
        """
        Outputs all snapshots names for global memory storage and/or for memory mapped files.
        These snapshots should be written there by the :py:meth:`onetick.py.Source.save_snapshot`.
        Output ticks have ``SNAPSHOT_NAME``, ``STORAGE_TYPE`` and ``DB_NAME`` fields.

        Parameters
        ----------
        snapshot_storage: 'memory' or 'memory_mapped_file'
            This parameter specifies the place of storage of the snapshot. Possible options are:

            * `memory` - the snapshot is stored in the dynamic (heap) memory of the process
              that ran (or is still running) the :py:meth:`onetick.py.Source.save_snapshot` for the snapshot.
            * `memory_mapped_file` - the snapshot is stored in a memory mapped file.
              For each symbol to get the location of the snapshot in the file system, ``ReadSnapshot`` looks at
              the **SAVE_SNAPSHOT_DIR** parameter value in the locator section for the database of the symbol.
            * `all` - shows both, `memory`, `memory_mapped_file` snapshots.

            Default: `all`

        See also
        --------
        | **SHOW_SNAPSHOT_LIST** OneTick event processor
        | :py:class:`onetick.py.ReadSnapshot`
        | :py:class:`onetick.py.FindSnapshotSymbols`
        | :py:meth:`onetick.py.Source.save_snapshot`
        | :py:meth:`onetick.py.Source.join_with_snapshot`

        Examples
        --------
        List snapshots from all snapshot storage types:

        >>> src = otp.ShowSnapshotList(snapshot_storage='all')  # doctest: +SKIP
        >>> otp.run(src)  # doctest: +SKIP
                Time SNAPSHOT_NAME        STORAGE_TYPE        DB_NAME
        0 2003-12-01    snapshot_1              MEMORY        DEMO_L1
        1 2003-12-01    snapshot_2              MEMORY        DEMO_L1
        2 2003-12-01    snapshot_3  MEMORY_MAPPED_FILE  SNAPSHOT_DEMO

        List snapshots from memory:

        >>> src = otp.ShowSnapshotList(snapshot_storage='memory')  # doctest: +SKIP
        >>> otp.run(src)  # doctest: +SKIP
                Time SNAPSHOT_NAME        STORAGE_TYPE        DB_NAME
        0 2003-12-01    snapshot_1              MEMORY        DEMO_L1
        1 2003-12-01    snapshot_2              MEMORY        DEMO_L1

        List snapshots from memory mapped files:

        >>> src = otp.ShowSnapshotList(snapshot_storage='memory_mapped_file')  # doctest: +SKIP
        >>> otp.run(src)  # doctest: +SKIP
                Time SNAPSHOT_NAME        STORAGE_TYPE        DB_NAME
        0 2003-12-01    snapshot_3  MEMORY_MAPPED_FILE  SNAPSHOT_DEMO
        """
        if 'schema' not in kwargs:
            kwargs['schema'] = {'SNAPSHOT_NAME': str, 'STORAGE_TYPE': str, 'DB_NAME': str}

        if self._try_default_constructor(**kwargs):
            return

        if not hasattr(otq, "ShowSnapshotList"):
            raise RuntimeError("Current version of OneTick don't support SHOW_SNAPSHOT_LIST EP")

        if snapshot_storage not in ['memory', 'memory_mapped_file', 'all']:
            raise ValueError('`snapshot_storage` must be one of "memory", "memory_mapped_file", "all"')

        super().__init__(
            _symbols=utils.adaptive,
            _base_ep_func=lambda: self.base_ep(snapshot_storage=snapshot_storage),
            **kwargs,
        )

    def base_ep(self, snapshot_storage='all'):
        snapshot_storage = snapshot_storage.upper()

        src = Source(otq.ShowSnapshotList(snapshot_storage=snapshot_storage))
        update_node_tick_type(src, otp.adaptive, otp.config.get('default_db', 'LOCAL'))

        return src


class FindSnapshotSymbols(Source):
    def __init__(
        self,
        snapshot_name='VALUE',
        snapshot_storage='memory',
        pattern='%',
        symbology=None,
        show_original_symbols=False,
        discard_on_match=False,
        db=None,
        tick_type=None,
        **kwargs,
    ):
        """
        Takes snapshot's name and storage type and produces a union of all symbols in the snapshot.
        It also optionally translates the symbols into a user-specified symbology and
        propagates those that match a specified name pattern as a tick with a single field: **SYMBOL_NAME**.

        If symbol name translation is performed, symbol names in native snapshot can optionally be propagated
        as another tick field: ``ORIGINAL_SYMBOL_NAME``, in which case symbols with missing translations
        are also propagated. Symbol name translation is performed as of the query symbol date and requires
        the reference database to be configured.

        The name of the database to query is extracted from the input symbol.
        For example, if an input symbol is ``TAQ::``, the symbols from the **TAQ** database will be returned.

        The ``pattern`` parameter can contain the special characters ``%`` (matches any number of characters) and
        ``_`` (matches any character).
        For example, the pattern ``I_M%`` returns any symbol beginning with I and has M as its third letter.

        If the ``discard_on_match`` parameter is set to ``True``, the names that do not match the pattern
        will be propagated.

        The timestamps of the ticks created by ``FindSnapshotSymbols`` are set to the start time of the query.

        Parameters
        ----------
        snapshot_name: str
            The name that was specified in :py:meth:`onetick.py.Source.save_snapshot` as a ``snapshot_name``
            during saving.

            Default: `VALUE`
        snapshot_storage: 'memory' or 'memory_mapped_file'
            This parameter specifies the place of storage of the snapshot. Possible options are:

            * `memory` - the snapshot is stored in the dynamic (heap) memory of the process
              that ran (or is still running) the :py:meth:`onetick.py.Source.save_snapshot` for the snapshot.
            * `memory_mapped_file` - the snapshot is stored in a memory mapped file.
              For each symbol to get the location of the snapshot in the file system, ``ReadSnapshot`` looks at
              the **SAVE_SNAPSHOT_DIR** parameter value in the locator section for the database of the symbol.

            Default: `memory`
        pattern: str
            The pattern for symbol selection. It can contain special characters ``%`` (matches any number of characters)
            and ``_`` (matches any character). To avoid this special interpretation of characters ``%`` and ``_``,
            the ``\\`` character must be added in front of them. Note that if symbol name translation is required,
            translated rather than original symbol names are checked to match the name pattern.

             Default: `%`
        symbology: Optional[str]
            The destination symbology for a symbol name translation, the latter being performed,
            if destination symbology is not empty and is different from that of the queried database.
        show_original_symbols: bool
            Switches original symbol name propagation as a tick field after symbol name translation is performed.
            Note that if this parameter is set to true, database symbols with missing translations are also propagated.

            Default: `False`
        discard_on_match: bool
            When set to true only ticks that did not match the filter are propagated,
            otherwise ticks that satisfy the filter condition are propagated.

            Default: `False`
        db: str, optional
            Database to use for snapshots search query.

            It's required to fill either this parameter or explicitly specify
            the ``symbols`` parameter of :py:func:`otp.run <onetick.py.run>`,
            or have bound symbols on any node of your query
        tick_type: str, optional
            Tick type.

        See also
        --------
        | **FIND_SNAPSHOT_SYMBOLS** OneTick event processor
        | :py:class:`onetick.py.ReadSnapshot`
        | :py:class:`onetick.py.ShowSnapshotList`
        | :py:meth:`onetick.py.Source.save_snapshot`
        | :py:meth:`onetick.py.Source.join_with_snapshot`

        Examples
        --------
        Find symbols for snapshot `some_snapshot` in database `S1`:

        >>> src = otp.FindSnapshotSymbols(snapshot_name='some_snapshot', db='S1')
        >>> otp.run(src, symbols='S1::')  # doctest: +SKIP
                Time SYMBOL_NAME
        0 2003-12-01    S1::AAPL
        1 2003-12-01    S1::AAAA
        2 2003-12-01    S1::MSFT

        Use ``pattern`` parameter to filter symbol names:

        >>> src = otp.FindSnapshotSymbols(snapshot_name='some_snapshot', db='S1', pattern='A%')
        >>> otp.run(src, symbols='S1::')  # doctest: +SKIP
                Time SYMBOL_NAME
        0 2003-12-01    S1::AAPL
        1 2003-12-01    S1::AAAA

        Select symbol names not matched by pattern:

        >>> src = otp.FindSnapshotSymbols(
        ...     snapshot_name='some_snapshot', db='S1', pattern='A%', discard_on_match=True,
        ... )
        >>> otp.run(src, symbols='S1::')  # doctest: +SKIP
                Time SYMBOL_NAME
        0 2003-12-01    S1::MSFT
        """
        if 'schema' not in kwargs:
            kwargs['schema'] = {'SYMBOL_NAME': str}

            if show_original_symbols:
                kwargs['schema'].update({'ORIGINAL_SYMBOL_NAME': str})

        if self._try_default_constructor(**kwargs):
            return

        if not hasattr(otq, "FindSnapshotSymbols"):
            raise RuntimeError("Current version of OneTick don't support SHOW_SNAPSHOT_LIST EP")

        if symbology is None:
            symbology = ''

        if snapshot_storage not in ['memory', 'memory_mapped_file']:
            raise ValueError('`snapshot_storage` must be one of "memory", "memory_mapped_file"')

        super().__init__(
            _base_ep_func=lambda: self.base_ep(
                snapshot_name=snapshot_name,
                snapshot_storage=snapshot_storage,
                pattern=pattern,
                symbology=symbology,
                show_original_symbols=show_original_symbols,
                discard_on_match=discard_on_match,
                db=db,
                tick_type=tick_type,
            ),
            **kwargs,
        )

    def base_ep(
        self,
        snapshot_name='VALUE',
        snapshot_storage='memory',
        pattern='%',
        symbology='',
        show_original_symbols=False,
        discard_on_match=False,
        db=None,
        tick_type=None,
    ):
        snapshot_storage = snapshot_storage.upper()

        src = Source(
            otq.FindSnapshotSymbols(
                snapshot_name=snapshot_name,
                snapshot_storage=snapshot_storage,
                pattern=pattern,
                symbology=symbology,
                show_original_symbols=show_original_symbols,
                discard_on_match=discard_on_match,
            )
        )
        update_node_tick_type(src, tick_type, db)

        return src
